"""Gaitrite data extraction."""
import logging
import os

import pandas as pd


logger = logging.getLogger('garjus.automations.etl_gaitrite')

# VUMC - all 4 walk types in a single excel file. column indicates walk type

# UPMC - file per walk type

# TODO: add Stride Velocity L/R mean and  Stride Velocity L/R stddev


def process(gaitrite_file):
    """Process Gaitrite file and return subset of data."""
    data = []

    if gaitrite_file.endswith('.zip'):
        logger.error('ignoring zip file')
        return []

    # Load each test, extract returns an array of tests.
    for d in _extract(gaitrite_file):
        data.append({
            'gaitrite_testrecord': str(d['Test Record #']),
            'gaitrite_datetime': str(d['Date / Time of Test']),
            'gaitrite_comments': str(d['Comments']),
            'gaitrite_velocity': str(d['Velocity']),
            'gaitrite_cadence': str(d['Cadence']),
            'gaitrite_steptime_left': str(d['Step Time(sec) L']),
            'gaitrite_steptime_right': str(d['Step Time(sec) R']),
            'gaitrite_stepextremity_left': str(d['Step Extremity(ratio) L']),
            'gaitrite_stepextremity_right': str(d['Step Extremity(ratio) R']),
            'gaitrite_stridelen_left': str(d['Stride Length(cm) L']),
            'gaitrite_stridelen_right': str(d['Stride Length(cm) R']),
            'gaitrite_swingtime_left': str(d['Swing Time(sec) L']),
            'gaitrite_swingtime_right': str(d['Swing Time(sec) R']),
            'gaitrite_stancetime_left': str(d['Stance Time(sec)  L']),
            'gaitrite_stancetime_right': str(d['Stance Time(sec)  R']),
            'gaitrite_funcambprofile': str(d['Functional Amb. Profile']),
            'gaitrite_normvelocity': str(d['Normalized Velocity  ']),
            'gaitrite_heeltime_left': str(d['Heel Off On Time L']),
            'gaitrite_heeltime_right': str(d['Heel Off On Time R'])
        })

    return data


def process_upmc(filename):
    """Process file and return subset of data."""

    # Load each test, extract returns an array of tests.
    d = _extract_upmc(filename)

    return {
        'gaitrite_datetime': str(d['Date / Time of Test']),
        'gaitrite_comments': str(d['Comments']),
        'gaitrite_velocity': str(d['Velocity']),
        'gaitrite_cadence': str(d['Cadence']),
        'gaitrite_steptime_left': str(d['Step Time(sec) L']),
        'gaitrite_steptime_right': str(d['Step Time(sec) R']),
        'gaitrite_stridelen_left': str(d['Stride Length(cm) L']),
        'gaitrite_stridelen_right': str(d['Stride Length(cm) R']),
        'gaitrite_swingtime_left': str(d['Swing Time(sec) L']),
        'gaitrite_swingtime_right': str(d['Swing Time(sec) R']),
        'gaitrite_stancetime_left': str(d['Stance Time(sec)  L']),
        'gaitrite_stancetime_right': str(d['Stance Time(sec)  R']),
    }


def _extract_upmc(filename):
    data = {}

    # First load from the top section of the file
    df = pd.read_excel(
        filename,
        header=None,
        usecols=[0,1],
        nrows=9,
        names=['FieldName', 'FieldValue'],

    )
    data['Comments'] = df[df['FieldName'] == 'Memo'].iloc[0].FieldValue
    data['Date / Time of Test'] = df[df['FieldName'] == 'Test Time'].iloc[0].FieldValue

    # Then load spreadsheet section
    df = pd.read_excel(filename, skiprows=11)
    df = df.fillna('')
    df = df.rename(columns={'Unnamed: 0': 'Measure', 'Unnamed: 1': 'LR'})

    try:
        data['Velocity'] = df[(df['Measure'] == 'Mean') & (df['LR'] == '')].iloc[0]['Velocity (cm./sec.)']
        data['Cadence'] = df[(df['Measure'] == 'Mean') & (df['LR'] == '')].iloc[0]['Cadence (steps/min.)']
        data['Step Time(sec) L'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Left')].iloc[0]['Step Time (sec.)']
        data['Step Time(sec) R'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Right')].iloc[0]['Step Time (sec.)']
        data['Stride Length(cm) L'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Left')].iloc[0]['Stride Length (cm.)']
        data['Stride Length(cm) R'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Right')].iloc[0]['Stride Length (cm.)']
        data['Swing Time(sec) L'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Left')].iloc[0]['Swing Time (sec.)']
        data['Swing Time(sec) R'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Right')].iloc[0]['Swing Time (sec.)']
        data['Stance Time(sec)  L'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Left')].iloc[0]['Stance Time (sec.)']
        data['Stance Time(sec)  R'] = df[(df['Measure'] == 'Mean') & (df['LR'] == 'Right')].iloc[0]['Stance Time (sec.)']
    except Exception as err:
        print(err)

    return data


def _extract(filename):
    """ Extract data from file that has a header row and one data row"""
    try:
        df = pd.read_csv(filename, dtype=str)
    except Exception:
        df = pd.read_excel(filename, dtype=str)

    try:
        df = df.dropna(subset=['Test Record #'])
    except Exception as err:
        logger.error(f'failed to extract gaitrite from excel:{err}')
        return []

    df = df.sort_values('Test Record #')

    # Fill nan with blanks
    df = df.fillna('')

    return df.to_records()


if __name__ == "__main__":
    import pprint

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    _dir = os.path.expanduser('~/Downloads')
    test_file = f'{_dir}/V1099_Gaitrite_Baseline.xlsx'

    data = process(test_file)
    pprint.pprint(data)



"""Garjus NDA image03 Management."""

# DESCRIPTION:
# image03 update and download, queries REDCap and XNAT to get updated info,
# save csv download each DIDOM zip
# ==========================================================================
# INPUTS:
# *project
# *start date (optional)
# *end date (optional)
# *mapping of scan types to "scan_type"
# ==========================================================================
# OUTPUTS:
# _image03.csv
# zip of each series organized as SESSION/SCAN/DICOM.zip
# ==========================================================================
# CSV columns are the 14 required fields (including conditionally required):
# subjectkey-->guid
# src_subject_id = internal subject number, e.g. 14295
# interview_date = scan date MM/DD/YYYY (NDA requires this format)
# interview_age = age at date in months-->use "dob" to calculate age
# sex
# image_file = (path to file on Box)
# image_description = (scan type)
# scan_type = (MR diffusion, fMRI, MR structural (T1), MR: FLAIR, etc.)
# scan_object = "Live"
# image_file_format = "DICOM"
# image_modality = "MRI"
# transformation_performed = "No"
# experiment_id = (fmri only, linked to experiment in NDA)
# bvek_bval_files = "Yes" (diffusion only)
# ==========================================================================

import os
import glob
import logging
from numpy import datetime_as_string

from zipfile import BadZipFile
import pandas as pd


logger = logging.getLogger('garjus.image03')

IMAGE03_TEMPLATE = "https://nda.nih.gov/api/datadictionary/v2/datastructure/image03/template"


def update(garjus, projects=None, startdate=None, enddate=None, sites=None):
    """Update image03 batches."""
    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating image03:{p}')
            _update_project(garjus, p, startdate, enddate, sites)


def download(garjus, project, image03_csv, download_dir):
    update_imagedir(garjus, project, image03_csv, download_dir)


def _parse_map(mapstring):
    """Parse map stored as string into dictionary."""

    parsed_map = mapstring.replace('=', ':')

    # Parse multiline string of delimited key value pairs into dictionary
    parsed_map = dict(x.strip().split(':', 1) for x in parsed_map.split('\n'))

    # Remove extra whitespace from keys and values
    parsed_map = {k.strip(): v.strip() for k, v in parsed_map.items()}

    return parsed_map


def _update_project(garjus, project, startdate=None, enddate=None, sites=None):
    xst2nei = None

    # Get map of Xnat scan types to NDA scan types
    xst2nst = garjus.project_setting(project, 'xst2nst')
    if not xst2nst:
        logger.debug('no xst2nst')
        return

    # Check for site specific experiment types in xst2nei
    if len(sites) == 1:
        site_data = garjus.sites(project)
        for rec in site_data:
            if rec['site_shortname'] == sites[0] and rec['site_xst2nei']:
                xst2nei = rec['site_xst2nei']

    if not xst2nei:
        # Get map of Xnat scan types to NDA experiment types
        xst2nei = garjus.project_setting(project, 'xst2nei')
        if not xst2nei:
            logger.debug('no xst2nei')
            return

    # Parse strings into dictionary
    xst2nst = _parse_map(xst2nst)
    xst2nei = _parse_map(xst2nei)

    logger.debug(f'settings:{project}:xst2nei={xst2nei}:xst2nst={xst2nst}')

    outfile = f'{project}_image03.csv'

    _make_image03_csv(
        garjus,
        project,
        xst2nst,
        xst2nei,
        outfile,
        startdate,
        enddate,
        sites=sites)


def _download_dicom_zip(scan, zipfile):
    dstdir = os.path.dirname(zipfile)

    # Make the output directory
    try:
        os.makedirs(dstdir)
    except FileExistsError:
        pass

    # Download zip of resource
    if scan.resource('DICOM').exists():
        try:
            dst_zip = scan.resource('DICOM').get(dstdir, extract=False)
            return dst_zip
        except BadZipFile as err:
            logger.error(f'error downloading:{err}')
            return None
    elif scan.resource('DICOMZIP').exists():
        try:
            res = scan.resource('DICOMZIP')
            src_zip = res.files().get()[0]
            res.file(src_zip).get(zipfile)
        except BadZipFile as err:
            logger.error(f'error downloading:{err}')
            return None
    else:
        logger.error(f'error downloading, DICOM not found')
        return None


def _touch_dicom_zip(scan, zipfile):
    dstdir = os.path.dirname(zipfile)

    # Make the output directory
    try:
        os.makedirs(dstdir)
    except FileExistsError:
        pass

    with open(zipfile, 'w'):
        pass


def _mr_info(scan_info, type_map, exp_map):
    scan_type = scan_info['SCANTYPE']
    scan_date = scan_info['DATE']
    subj_label = scan_info['SUBJECT']
    scan_label = scan_info['SCANID']

    zip_path = os.path.join(
        f'{subj_label}_MR_{str(scan_date).split(" ")[0]}'.replace('-',''),
        '{}_{}'.format(scan_label, scan_type.replace(' ', '_')),
        'DICOM.zip')

    info = {
        'scan_object': 'Live',
        'image_file_format': 'DICOM',
        'image_modality': 'MRI',
        'transformation_performed': 'No'}
    info['image_file'] = zip_path
    info['src_subject_id'] = subj_label
    info['interview_date'] = scan_date
    info['image_description'] = scan_type
    info['scan_type'] = type_map[scan_type]

    if 'SEX' in scan_info:
        info['sex'] = scan_info['SEX']
    else:
        logger.debug(f'SEX not found')
        info['sex'] = ''

    if 'GUID' in scan_info:
        info['subjectkey'] = scan_info['GUID']
    else:
        logger.debug(f'GUID not found')
        info['subjectkey'] = ''

    if 'SCANAGE' in scan_info:
        info['interview_age'] = scan_info['SCANAGE']
    else:
        logger.debug(f'SCANAGE not found')
        info['interview_age'] = ''

    if scan_type.startswith('DTI'):
        info['bvek_bval_files'] = 'Yes'

    if scan_type in exp_map.keys():
        info['experiment_id'] = exp_map[scan_type]

    return info


def _pet_info(scan_info, type_map):
    scan_type = scan_info['SCANTYPE']
    scan_date = scan_info['DATE']
    subj_label = scan_info['SUBJECT']
    scan_label = scan_info['SCANID']
    zip_path = os.path.join(
        f'{subj_label}_PET_{str(scan_date).split(" ")[0]}'.replace('-',''),
        '{}_{}'.format(scan_label, scan_type.replace(' ', '_')),
        'DICOM.zip')

    info = {
        'scan_object': 'Live',
        'image_file_format': 'DICOM',
        'image_modality': 'PET',
        'transformation_performed': 'No'}

    info['image_file'] = zip_path
    info['src_subject_id'] = subj_label
    info['interview_date'] = scan_date
    info['image_description'] = scan_type
    info['scan_type'] = type_map[scan_type]

    if 'SEX' in scan_info:
        info['sex'] = scan_info['SEX']
    else:
        logger.debug(f'SEX not found')
        info['sex'] = ''

    if 'GUID' in scan_info:
        info['subjectkey'] = scan_info['GUID']
    else:
        logger.debug(f'GUID not found')
        info['subjectkey'] = ''

    if 'SCANAGE' in scan_info:
        info['interview_age'] = scan_info['SCANAGE']
    else:
        logger.debug(f'SCANAGE not found')
        info['interview_age'] = ''

    return info


def get_image03_df(mr_scans, pet_scans, type_map, exp_map):
    data = []

    # Load the MRIs
    for cur_scan in mr_scans.to_dict('records'):
        data.append(_mr_info(cur_scan, type_map, exp_map))

    # Load the PETs
    for cur_scan in pet_scans.to_dict('records'):
        data.append(_pet_info(cur_scan, type_map))

    # Initialize with template columns, ignoring first row
    logger.debug('load template from web')
    df = pd.read_csv(IMAGE03_TEMPLATE, skiprows=1)

    # Append our records
    if len(data) > 0:
        if len(df) > 0:
            df = pd.concat([df, pd.DataFrame(data)])
        else:
            df = pd.DataFrame(data, columns=df.columns)

    return df


def update_files(garjus, project, df, download_dir):
    ecount = 0
    dcount = 0

    # Merge in xnat info
    scans = garjus.scans(projects=[project])
    sessions = scans[['SUBJECT', 'SESSION', 'DATE']].drop_duplicates()
    sessions['interview_date'] = pd.to_datetime(sessions['DATE']).dt.strftime('%m/%d/%Y')

    df = pd.merge(
        df,
        sessions,
        how='left',
        left_on=['src_subject_id', 'interview_date'],
        right_on=['SUBJECT', 'interview_date'])

    with garjus.xnat() as xnat:
        for i, f in df.iterrows():
            # Determine scan label
            scan_label = f['image_file'].split('/')[1].split('_')[0]

            # Local file path
            cur_file = os.path.join(download_dir, f['image_file'])
            if os.path.exists(cur_file):
                ecount += 1
                continue

            # connect to scan
            scan = xnat.select_scan(
                project,
                f['src_subject_id'],
                f['SESSION'],
                scan_label)

            # get the file
            logger.info(f'downloading:{cur_file}')
            _download_dicom_zip(scan, cur_file)
            dcount += 1

    logger.info(f'{ecount} existing files, {dcount} downloaded files')


def same_data(filename, df):
    is_same = False

    # Load data from the file
    df2 = pd.read_csv(filename, dtype=str, skiprows=1)

    # Compare contents
    try:
        if len(df.compare(df2)) == 0:
            is_same = True
    except ValueError:
        pass

    # Return the result
    logger.info(f'is_same={is_same}')
    return is_same


def not_downloaded(df, image_dir):
    if not os.path.isdir(image_dir):
        return df

    # Get list of DICOM zips already existing
    zip_list = glob.glob(f'{image_dir}/*/*/*/DICOM.zip')

    # Standardize naming
    zip_list = ['/'.join(z.rsplit('/', 4)[2:4]).upper() for z in zip_list]

    # Now only include not downloaded
    df = df[~df.image_file.str.rsplit('/', n=4).str[2:4].apply('/'.join).str.upper().isin(zip_list)]

    return df


def _make_image03_csv(
    garjus,
    project,
    type_map,
    exp_map,
    outfile,
    startdate=None,
    enddate=None,
    sites=None,
):
    dfs = garjus.subjects(project, include_dob=True)

    # Get the MRIs
    mscans = garjus.scans(
        projects=[project],
        scantypes=type_map.keys(),
        modalities=['MR'],
        sites=sites,
        startdate=startdate,
        enddate=enddate)

    # Get the PETs
    pscans = garjus.scans(
        projects=[project],
        scantypes=type_map.keys(),
        modalities=['PET'],
        sites=sites,
        startdate=startdate,
        enddate=enddate)

    # merge in subject data
    mscans = pd.merge(mscans, dfs, left_on='SUBJECT', right_on='ID')
    pscans = pd.merge(pscans, dfs, left_on='SUBJECT', right_on='ID')

    if 'DOB' in mscans:
        mscans.DOB = pd.to_datetime(mscans.DOB)
        pscans.DOB = pd.to_datetime(pscans.DOB)

        # Calculate Scan age in integer of months as a string
        mscans['SCANAGE'] = (mscans['DATE'] + pd.DateOffset(days=15)) - mscans['DOB']
        mscans['SCANAGE'] = mscans['SCANAGE'].values.astype('<m8[M]').astype('int').astype('str') 
        pscans['SCANAGE'] = (pscans['DATE'] + pd.DateOffset(days=15)) - pscans['DOB']
        pscans['SCANAGE'] = pscans['SCANAGE'].values.astype('<m8[M]').astype('int').astype('str')
    else:
        logger.debug(f'DOB not found, cannot calculate scan age')
        mscans['SCANAGE'] = ''
        pscans['SCANAGE'] = ''

    # get the image03 formatted
    df = get_image03_df(
        mscans,
        pscans,
        type_map,
        exp_map)

    # Set columns to same list and order as NDA template
    df = df[pd.read_csv(IMAGE03_TEMPLATE, skiprows=1).columns]

    # Compare to existing csv and only write to new file if something changed
    if not os.path.exists(outfile) or not same_data(outfile, df):
        # Save data to file
        logger.info(f'saving to csv file:{outfile}')

        # write the header
        with open(outfile, 'w') as fp:
            fp.write('"image","03"\n')

        # write the rest
        df.to_csv(outfile, mode='a', index=False, date_format='%m/%d/%Y')
    else:
        logger.info(f'no new data, use existing csv:{outfile}')


def make_dirs(dir_path):
    logger.debug(f'make_dirs{dir_path}')
    try:
        os.makedirs(dir_path)
    except OSError:
        if not os.path.isdir(dir_path):
            raise


def update_imagedir(garjus, project, csvfile, download_dir):

    df = pd.read_csv(csvfile, dtype=str, skiprows=1)

    # Remove already downloaded
    df = not_downloaded(df, download_dir)

    # Update DICOM zips
    update_files(garjus, project, df, download_dir)



import os
import datetime

import dash
import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template


dbc_css = "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"


# TODO: fix this to find path more dynamically
assets_path = os.path.expanduser('~/git/garjus/garjus/dashboard/assets')

hour = datetime.datetime.now().hour

darkmode = True

#$if hour < 9 or hour > 17:
if darkmode:
    stylesheets = [dbc.themes.DARKLY, dbc_css]
    load_figure_template("darkly")
else:
    stylesheets = [dbc.themes.FLATLY, dbc_css]
    load_figure_template("flatly")

app = dash.Dash(
    __name__,
    external_stylesheets=stylesheets,
    assets_folder=assets_path,
)

server = app.server

#app.config.suppress_callback_exceptions = True

# more here:
# https://hellodash.pythonanywhere.com
# https://codepen.io/chriddyp/pen/bWLwgP.css
# dbc.themes.DARKLY flat dark
# dbc.themes.FLATLY flat light
# dbc.themes.LUMEN bright
# dbc.themes.SLATE dark
# dbc.themes.SPACELAB buttony
# dbc.themes.YETI nice



import glob
import re
import logging

from ..utils_redcap import upload_file


logger = logging.getLogger('garjus.automations.edat_limbo2redcap')


def process_project(
    project,
    events,
    raw_field,
    tab_field,
    edat_prefix,
    limbo_dir,
    event2sess=None,
):
    results = []
    def_field = project.def_field
    fields = [def_field, raw_field, tab_field]
    id2subj = {}

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get mri records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event = r['redcap_event_name']
        sess_num = '1'

        if event2sess is not None:
            sess_num = event2sess[event]

        try:
            subj = id2subj[record_id]
            logger.debug(f'{subj}:{event} using secondary id:{record_id}')
        except KeyError as err:
            logger.debug(f'record without subject number:{err}')
            continue

        if not subj:
            logger.warn(f'blank subject number:{record_id}')
            continue

        # Check for no raw file
        if r[raw_field]:
            logger.debug(f'already uploaded:{subj}:{event}')
            continue

        # Check for existing converted file
        if r[tab_field]:
            logger.debug(f'already converted:{subj}:{event}')
            continue

        # Find a file for this record
        logger.debug(f'looking for files:{subj}:{event}')

        # Get just the numeric portion of the subject number
        subj_num = re.sub(r'[^0-9]', '', subj)

        # Find files for this subject
        sess_glob = f'{limbo_dir}/{edat_prefix}-{subj_num}-{sess_num}.edat?'
        file_list = sorted(glob.glob(sess_glob))
        file_count = len(file_list)
        if file_count <= 0:
            logger.debug(f'no file found:{subj}:{event}:{sess_glob}')
            continue
        elif file_count > 1:
            logger.debug(f'too many files:{subj}:{event}:{sess_glob}')
            continue

        # Upload file to redcap
        edat_file = file_list[0]
        logger.debug(f'uploading file:{edat_file}')
        try:
            result = upload_file(
                project,
                record_id,
                raw_field,
                edat_file,
                event_id=event)

            logger.debug(f'uploaded:{subj}:{event}:{edat_file}')
        except (ValueError) as err:
            logger.error(f'error uploading:{edat_file}:{err}')

        if not result:
            logger.error(f'upload failed:{subj}:{event}')
            continue

        logger.debug(f'uploaded:{subj}:{event}:{raw_field}')
        results.append({
            'result': 'COMPLETE',
            'description': 'edat_limbo2redcap',
            'category': 'edat_limbo2redcap',
            'subject': subj,
            'session': '',
            'scan': '',
            'event': event,
            'field': raw_field})

    return results



import pandas as pd
from garjus import Garjus
from garjus.utils_redcap import field2events


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        print(err)
        return False


def _extract(filename):
    # Load file data
    df = pd.read_csv(filename)
    return df


def _transform(rc, df):
    subj2id = {}
    data = []
    def_field = rc.def_field
    sec_field = None

    # Load the subject numbers
    sec_field = rc.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = rc.export_records(fields=[def_field, sec_field])
        subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
    else:
        rec = rc.export_records(fields=[def_field])
        subj2id = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    for i, row in df.iterrows():
        record_id = subj2id[row['subject']]
        data.append(
            {
                rc.def_field: record_id,
                'fallypride_accumbens': str(row['accumbens']),
                'fallypride_amygdala': str(row['amygdala']),
                'fallypride_caudate': str(row['caudate']),
                'fallypride_pallidum': str(row['pallidum']),
                'fallypride_putamen': str(row['putamen']),
                'fallypride_thalamus': str(row['thalamus']),
                'fallypride_complete': '2',
            }
        )

    # Get existing records
    _fields = ['fallypride_amygdala']
    _events = field2events(rc, 'fallypride_amygdala')
    records = rc.export_records(fields=_fields)
    records = [x for x in records if x['redcap_event_name'] in _events]

    existing = [x[def_field] for x in records if x.get('fallypride_amygdala', False)]

    data = [x for x in data if x[def_field] not in existing]

    # Copy event name from existing records
    for d in data:
        for r in records:
            if d[def_field] == r[def_field]:
                d['redcap_event_name'] = r['redcap_event_name']

    return data


if __name__ == "__main__":
    import sys

    filename = sys.argv[1]

    print(filename)

    # Get the redcap via garjus
    g = Garjus()
    rc = g.primary('D3')

    # Extract from file as pandas dataframe
    df = _extract(filename)

    # Prep for redcap by converting to records with field names that match redcap
    data = _transform(rc, df)

    # Upload to redcap
    print(f'loading:n={len(data)} records')
    _load(rc, data)
    print('done!')



import logging
import os
from datetime import datetime
import pandas as pd

from ....garjus import Garjus


logger = logging.getLogger('dashboard.stats.data')


def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/statsdata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def run_refresh(filename, projects):
    # force a requery
    df = get_data(projects)

    save_data(df, filename)

    return df


def load_options(selected_proj=None):

    try:
        garjus = Garjus()

        if not garjus.redcap_enabled():
            return [], []

        proj_options = garjus.projects()

        proc_options = []
        if selected_proj:
            for p in selected_proj:
                proc_options.extend(garjus.stattypes(p))

        proc_options = sorted(list(set(proc_options)))

        return proj_options, proc_options
    except Exception as err:
        logger.debug(err)
        return [], []


def load_data(projects, refresh=False, filename=None):
    if not filename:
        filename = get_filename()

    if refresh or not os.path.exists(filename):
        # TODO: check for old file and refresh too
        run_refresh(filename, projects)

    logger.info('reading data from file:{}'.format(filename))
    return read_data(filename)


def read_data(filename):
    df = pd.read_pickle(filename)
    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def get_data(projects):
    df = pd.DataFrame()
    garjus = Garjus()

    if not projects:
        return df

    # Concat project stats list of stats
    assessors = garjus.assessors(projects)
    assessors = pd.concat([assessors, garjus.subject_assessors(projects=projects)])

    for p in sorted(projects):
        # Load stats
        stats = garjus.stats(p, assessors)
        df = pd.concat([df, stats])

    # Apply tweaks
    if 'SESSTYPE' in df.columns:
        df['SESSTYPE'] = df['SESSTYPE'].fillna('UNKNOWN')
    else:
        df['SESSTYPE'] = 'UNKNOWN'

    if 'SITE' in df.columns:
        df['SITE'] = df['SITE'].fillna('UNKNOWN')
    else:
        df['SITE'] = 'UNKNOWN'

    if 'DATE' in df:
        # Format as string to drop the time portion
        df['DATE'] = df['DATE'].dt.strftime('%Y-%m-%d')

    # http link to session in xnat
    df['SESSIONLINK'] = garjus.xnat().host + \
        '/data/projects/' + df['PROJECT'] + \
        '/subjects/' + df['SUBJECT'] + \
        '/experiments/' + df['SESSION']

    return df


def filter_data(df, proctypes, timeframe, sesstypes):
    if df.empty:
        # It's already empty, just send it back
        return df

    logger.debug(f'applying filters:{proctypes}:{timeframe}:{sesstypes}')

    # Filter by proctype
    if proctypes:
        logger.debug(f'filtering by proctypes:{proctypes}')
        df = df[df['PROCTYPE'].isin(proctypes)]
    else:
        logging.debug('no proctypes')
        df = df[df['PROCTYPE'].isin([])]

    # Filter by timeframe
    if timeframe in ['1day', '7day', '30day', '365day']:
        logging.debug('filtering by ' + timeframe)
        then_datetime = datetime.now() - pd.to_timedelta(timeframe)
        df = df[pd.to_datetime(df.DATE) > then_datetime]
    else:
        # ALL
        logging.debug('not filtering by time')
        pass

    # Filter by sesstype
    if sesstypes:
        logging.debug(f'filtering by sesstypes:{sesstypes}')
        df = df[df['SESSTYPE'].isin(sesstypes)]
    else:
        logging.debug('no sesstypes')

    # Remove empty columns
    df = df.dropna(axis=1, how='all')

    return df



import logging
import sys
import importlib
import tempfile

from garjus.utils_redcap import download_file, field2events


logger = logging.getLogger('test_gaitrite')


def run(project):
    data = []
    results = []
    events = []
    records = []
    file_field = 'gaitrite_upload'
    done_field = 'gaitrite_testrecord'

    # load the automation
    try:
        mod = importlib.import_module(f'garjus.automations.etl_gaitrite')
    except ModuleNotFoundError as err:
        logger.error(f'error loading module:{err}')
        return

    # records will include both the file record and the repeating records
    records = project.export_records(
        fields=[file_field, done_field],
        events=field2events(project, file_field)
    )

    # Get unique list of record/event
    recordevents = list(set(
        [(x[project.def_field], x['redcap_event_name']) for x in records]))

    # Initialize counts
    counts = {x: 0 for x in recordevents}

    # Count repeats
    repeats = [x for x in records if x['redcap_repeat_instance']]

    # Count total records for each id/event
    for r in repeats:
        record_id = r[project.def_field]
        event_id = r['redcap_event_name']
        counts[(record_id, event_id)] += 1

    # Process each record
    for r in [x for x in records if not x['redcap_repeat_instance']]:
        data = []
        record_id = r[project.def_field]
        event_id = r['redcap_event_name']

        if not r[file_field]:
            logger.debug(f'no data file:{record_id}:{event_id}')
            continue

        if counts[(record_id, event_id)] > 0:
            logger.debug(f'already done:{record_id}:{event_id}')
            continue

        logger.debug(f'etl_gaitrite:{record_id}:{event_id}')

        filename = r[file_field]

        if filename.endswith('.zip'):
            logger.debug('ignoring zip')
            continue

        with tempfile.TemporaryDirectory() as tmpdir:
            data_file = f'{tmpdir}/{filename}'

            try:
                # Download files from redcap
                logger.debug(f'downloading VUMC file:{data_file}')
                download_file(
                    project,
                    record_id,
                    file_field,
                    data_file,
                    event_id=event_id
                )

                # Process downloaded file to extract data
                data = mod.process(data_file)

            except Exception as err:
                logger.error(err)
                continue

        # Load data back to redcap
        results.append({'subject': record_id, 'event': event_id, 'field': 'gaitrite_upload'})
        for d in data:
            d[project.def_field] = record_id
            d['redcap_event_name'] = event_id
            d['redcap_repeat_instrument'] = 'gaitrite'
            d['redcap_repeat_instance'] = 'new'
            d['gaitrite_complete'] = '2'

        if len(data) > 0:
            try:
                response = project.import_records(data)
                assert 'count' in response
                logger.debug(f'uploaded:{record_id}:{event_id}')
            except AssertionError as e:
                logger.error('error uploading', record_id, e)

    return results


def run_upmc(project):
    data = []
    results = []
    events = []
    records = []
    file_field = 'gaitrite_file'
    done_field = 'gaitrite_comments'

    # load the automation
    try:
        mod = importlib.import_module(f'garjus.automations.etl_gaitrite')
    except ModuleNotFoundError as err:
        logger.error(f'error loading module:{err}')
        return

    records = project.export_records(
        fields=[file_field, done_field],
    )

    records = [x for x in records if x['redcap_repeat_instrument'] == 'gaitrite']

    # Process each record
    for r in records:
        data = []
        record_id = r[project.def_field]
        repeat_id = r['redcap_repeat_instance']
        event_id = r['redcap_event_name']

        if r[done_field]:
            logger.debug(f'already done:{record_id}:{event_id}:{repeat_id}')
            continue

        if not r[file_field]:
            logger.debug(f'no data file:{record_id}:{event_id}:{repeat_id}')
            continue


        logger.debug(f'etl_gaitrite:{record_id}:{event_id}:{repeat_id}')

        with tempfile.TemporaryDirectory() as tmpdir:
            data_file = f'{tmpdir}/{r.get(file_field)}'

            try:
                # Download file from redcap
                logger.debug(f'downloading UPMC file:{data_file}')
                download_file(
                    project,
                    record_id,
                    file_field,
                    data_file,
                    event_id=event_id,
                    repeat_id=repeat_id,
                )

                # Process downloaded file to extract data
                d = mod.process_upmc(data_file)

                try:
                    # Complete the record with identifiers
                    d[project.def_field] = record_id
                    d['redcap_event_name'] = event_id
                    d['redcap_repeat_instrument'] = 'gaitrite'
                    d['redcap_repeat_instance'] = str(repeat_id)
                    d['gaitrite_complete'] = '2'
                except Exception as err:
                    logger.error(err)

                # Add to list for redcap upload
                data.append(d)

                # Save history
                results.append({'subject': record_id, 'event': event_id, 'field': file_field})

            except Exception as err:
                logger.error(err)
                continue

        if len(data) > 0:
            # Load data back to redcap
            try:
                response = project.import_records(data)
                assert 'count' in response
                logger.debug(f'uploaded:{record_id}:{event_id}')
            except AssertionError as e:
                logger.error('error uploading', record_id, e)

    return results


if __name__ == "__main__":
    import redcap
    from garjus.automations.etl_gaitrite import process

    api_url = 'https://redcap.vanderbilt.edu/api/'
    api_key = sys.argv[1]
    rc = redcap.Project(api_url, api_key)

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    results = run(rc)
    results += run_upmc(rc)
    print(results)
    print('Done!')



"""Tasks."""
import logging

from .processors import build_processor


logger = logging.getLogger('garjus.tasks')


def update(garjus, projects=None, types=None):
    """Update tasks."""
    if not garjus.xnat_enabled():
        logger.debug('no xnat, cannot update tasks')
        return

    for p in (projects or garjus.projects()):
        if p in projects:
            logger.debug(f'updating tasks:{p}')
            _update_project(garjus, p, types=types)


def _update_project(garjus, project, types=None):
    # Get protocol data, download yaml files as needed
    protocols = garjus.processing_protocols(project, download=True)

    if len(protocols) == 0:
        logger.info(f'no processing protocols for project:{project}')
        return

    if types:
        protocols = protocols[protocols.TYPE.isin(types)]

    # Get scan/assr/sgp data
    assessors = garjus.assessors(projects=[project])
    scans = garjus.scans(projects=[project])
    sgp = garjus.subject_assessors(projects=[project])

    project_data = {}
    project_data['name'] = project
    project_data['scans'] = scans
    project_data['assessors'] = assessors
    project_data['sgp'] = sgp

    # Iterate processing protocols
    for i, row in protocols.iterrows():
        filepath = row['FILE']

        logger.info(f'file:{project}:{filepath}')

        user_inputs = row.get('ARGS', None)
        if user_inputs:
            logger.debug(f'overrides:{user_inputs}')
            rlist = user_inputs.strip().split('\r\n')
            rdict = {}
            for arg in rlist:
                try:
                    key, val = arg.split(':', 1)
                    rdict[key] = val.strip()
                except ValueError as e:
                    msg = f'invalid arguments:{project}:{filepath}:{arg}:{e}'
                    raise Exception(msg)

            user_inputs = rdict
            logger.debug(f'user_inputs:{user_inputs}')

        if row['FILTER']:
            include_filters = str(row['FILTER']).replace(' ', '').split(',')
        else:
            include_filters = []

        build_processor(
            garjus,
            filepath,
            user_inputs,
            project_data,
            include_filters)



import click
import pprint
import logging
import os

from .garjus import Garjus


logging.basicConfig(
    format='%(asctime)s - %(levelname)s:%(name)s:%(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.option('--quiet/--no-quiet', default=False)
def cli(debug, quiet):
    if debug:
        click.echo('garjus! debug')
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger('werkzeug').setLevel(logging.DEBUG)
        logging.getLogger('dash').setLevel(logging.DEBUG)

    if quiet:
        logging.getLogger().setLevel(logging.ERROR)
        logging.getLogger('werkzeug').setLevel(logging.ERROR)
        logging.getLogger('dash').setLevel(logging.ERROR)


@cli.command('copysess')
@click.argument('src', required=True)
@click.argument('dst', required=True)
def copy_session(src, dst):
    click.echo('garjus! copy session')
    Garjus().copy_sess(src, dst)


@cli.command('copyscan')
@click.argument('src', required=True)
@click.argument('dst', required=True)
def copy_scan(src, dst):
    click.echo('garjus! copy scan')
    Garjus().copy_scan(src, dst)


@cli.command('setsesstype')
@click.argument('src', required=True)
@click.argument('sesstype', required=True)
def set_sesstype(src, sesstype):
    click.echo('garjus! set session type')
    Garjus().set_session_type(src, sesstype)


@cli.command('setsite')
@click.argument('src', required=True)
@click.argument('site', required=True)
def set_site(src, site):
    click.echo('garjus! set session site')
    Garjus().set_session_site(src, site)


@cli.command('issues')
@click.option('--project', '-p', 'project')
@click.pass_context
def issues(ctx, project):
    click.echo('garjus! issues')
    g = Garjus()
    pprint.pprint(g.issues(project))


@cli.command('subjects')
@click.option('--project', '-p', 'project')
@click.option('--csv', '-c', 'csv', required=False)
@click.pass_context
def subjects(ctx, project, csv):
    click.echo('garjus! subjects')
    g = Garjus()
    import pandas as pd
    pd.set_option('display.max_rows', None)
    subjects = g.subjects(project)
    if csv:
        subjects = subjects.set_index('ID')
        columns = subjects.columns
        columns = [x for x in columns if x not in ['identifier_id', 'GUID']]
        subjects[columns].to_csv(csv)
    else:
        pprint.pprint(subjects)


@cli.command('orphans')
@click.option('--project', '-p', 'project')
@click.option('--delete/--no-delete', default=False)
def orphans(project, delete):
    click.echo('garjus! orphans')
    g = Garjus()

    orphans = g.orphans(project)

    print(*orphans, sep='\n')

    if delete:
        print('deleting')
        for cur in orphans:
            print('delete', cur)
            g.delete_assessor(project, cur)
    else:
        print('not deleting')


@cli.command('activity')
@click.option('--project', '-p', 'project')
def activity(project):
    click.echo('garjus! activity')
    g = Garjus()
    pprint.pprint(g.activity(project))


@cli.command('analyses')
@click.option('--project', '-p', 'project')
def analyses(project):
    click.echo('garjus! analyses')
    g = Garjus()
    pprint.pprint(g.analyses(project))


@cli.command('getinputs')
@click.argument('analysis_id', required=True)
@click.argument('download_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
@click.option('--processor', '-y', 'processor', required=False)
def getinputs(project, analysis_id, download_dir, processor):
    click.echo('garjus! getinputs')
    g = Garjus()
    g.get_analysis_inputs(project, analysis_id, download_dir, processor)


@cli.command('getoutputs')
@click.argument('analysis_id', required=True)
@click.argument('download_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
def getoutputs(project, analysis_id, download_dir):
    click.echo('garjus! getoutputs')
    g = Garjus()
    g.get_analysis_outputs(project, analysis_id, download_dir)


@cli.command('download')
@click.argument('download_dir', required=True)
@click.option('--type', '-t', 'proctype', multiple=False, required=True)
@click.option('--resource', '-r', 'resource', multiple=True, required=True)
@click.option('--file', '-f', 'file', multiple=True, required=False)
@click.option('--project', '-p', 'project', required=True)
@click.option('--sesstype', '-s', 'sesstype', multiple=True, required=False)
@click.option('--analysis', '-a', 'analysis', required=False)
@click.option('--session', '-e', 'session', multiple=True, required=False)
@click.option(
    "--scan", is_flag=True, default=False, help="scans instead of assessors.")
def download(
    project,
    proctype,
    download_dir,
    resource,
    file,
    scan,
    sesstype,
    analysis,
    session
):
    click.echo('garjus! download')
    g = Garjus()
    if scan:
        g.download_scantype(
            project,
            download_dir,
            proctype,
            resource,
            file,
            sesstype,
            session
        )
    else:
        g.download_proctype(
            project,
            download_dir,
            proctype,
            resource,
            file,
            sesstype,
            analysis,
            session
        )


@cli.command('switch')
@click.option('--type', '-t', 'proctype', multiple=False, required=True)
@click.option('--old', '-o', 'oldstatus', multiple=False, required=True)
@click.option('--new', '-n', 'newstatus', multiple=False, required=True)
@click.option('--project', '-p', 'project', required=True)
@click.option('--sesstype', '-s', 'sesstype', multiple=True, required=False)
@click.option('--session', '-e', 'session', multiple=True, required=False)
def switch_status(
    project,
    proctype,
    oldstatus,
    newstatus,
    sesstype,
    session
):
    click.echo('garjus! switch status')
    Garjus().switch_status(
        project,
        proctype,
        oldstatus,
        newstatus,
        sesstype,
        session
    )


@cli.command('run')
@click.option('--project', '-p', 'project', required=True)
@click.option('--subjects', '-s', 'subjects', required=False, multiple=True)
@click.option('--repo', '-r', 'repo', required=True)
@click.option('--dir', '-d', 'jobdir', type=click.Path(), required=True)
@click.option('--csv', '-c', 'csv', required=False)
@click.option('--yaml', '-y', 'yamlfile', type=click.Path(), required=False)
def run(project, subjects, repo, jobdir, csv, yamlfile):
    click.echo('garjus! run')

    g = Garjus()
    g.run_analysis(project, subjects, repo, jobdir, csv, yamlfile)


@cli.command('finish')
@click.argument('analysis_id', required=True)
@click.argument('analysis_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
@click.option('--processor', '-y', 'processor', required=False)
def finish(project, analysis_id, analysis_dir, processor):
    click.echo('garjus! finish')
    g = Garjus()
    g.finish_analysis(project, analysis_id, analysis_dir, processor)


@cli.command('tasks')
def tasks():
    click.echo('garjus! tasks')
    g = Garjus()
    pprint.pprint(g.tasks())


@cli.command('update')
@click.argument(
    'choice',
    type=click.Choice([
        'stats',
        'issues',
        'progress',
        'automations',
        'compare',
        'tasks',
        'analyses',
        'scans'
    ]),
    required=False,
    nargs=-1)
@click.option('--project', '-p', 'project', multiple=True)
@click.option('--types', '-t', 'types', multiple=True, required=False)
def update(choice, project, types):
    click.echo('garjus! update')
    g = Garjus()
    g.update(projects=project, choices=choice, types=types)
    click.echo('ALL DONE!')


@cli.command('progress')
@click.option('--project', '-p', 'project')
def progress(project):
    click.echo('garjus! progress')
    if project:
        project = project.split(',')

    g = Garjus()
    pprint.pprint(g.progress(projects=project))


@cli.command('processing')
@click.option('--project', '-p', 'project', required=True)
def processing(project):
    click.echo('garjus! processing')

    g = Garjus()
    pprint.pprint(g.processing_protocols(project))


@cli.command('report')
@click.option('--project', '-p', 'project', required=True)
@click.option('--monthly/--no-monthly', default=True, required=False)
def report(project, monthly):
    click.echo('garjus! report')
    Garjus().report(project, monthly=monthly)


@cli.command('stats')
@click.option('--projects', '-p', 'projects', required=True)
@click.option('--types', '-t', 'proctypes', required=False)
@click.option('--sesstypes', '-s', 'sesstypes', required=False)
@click.option('--analysis', '-a', 'analysis', required=False)
@click.option('--persubject', is_flag=True)
@click.option('--sessions', '-e', 'sessions', required=False)
@click.argument('csv', required=True)
def stats(projects, proctypes, sesstypes, csv, persubject, analysis, sessions):
    click.echo('garjus! stats')
    Garjus().export_stats(
        projects,
        proctypes,
        sesstypes,
        csv,
        persubject,
        analysis,
        sessions)


@cli.command('export')
@click.option('--projects', '-p', 'projects', required=True)
@click.option('--types', '-t', 'proctypes', required=False)
@click.option('--sesstypes', '-s', 'sesstypes', required=False)
@click.option('--analysis', '-a', 'analysis', required=False)
@click.option('--sessions', '-e', 'sessions', required=False)
@click.argument('filename', required=True)
def export(filename, projects, proctypes, sesstypes, analysis, sessions):
    click.echo('garjus! export')
    Garjus().export_zip(
        filename,
        projects,
        proctypes,
        sesstypes,
        analysis,
        sessions)


@cli.command('statshot')
@click.option('--projects', '-p', 'projects', required=True, multiple=True)
@click.option('--proctypes', '-t', 'proctypes', required=False)
@click.option('--sesstypes', '-s', 'sesstypes', required=False)
@click.option('--exclude', '-x', 'exclude', required=False)
@click.option('--guid/--no-guid', default=False)
@click.option('--ident/--no-ident', default=False)
def statshot(projects, proctypes, sesstypes, exclude, guid, ident):
    click.echo('garjus! statshot')

    # Split projects into lists    
    projects = [x.split(',') for x in projects]

    # Flatten to single list
    projects = sum(projects, [])

    Garjus().statshot(
        projects,
        proctypes,
        sesstypes,
        exclude,
        guid=guid,
        ident=ident)

@cli.command('compare')
@click.option('--project', '-p', 'project', required=True)
def compare(project):
    click.echo('garjus! compare')
    Garjus().compare(project)


@cli.command('importdicom')
@click.argument('src', required=True)
@click.argument('dst', required=True)
def import_dicom(src, dst):
    click.echo('garjus! import')
    g = Garjus()
    g.import_dicom(src, dst)


@cli.command('importnifti')
@click.argument('src', required=True)
@click.argument('dst', required=True)
@click.option('--modality', '-m', 'modality', required=False)
def import_nifti(src, dst, modality):
    click.echo('garjus! importnifti')
    g = Garjus()
    g.import_nifti(src, dst, modality=modality)


@cli.command('pdf')
@click.argument('src', required=True)
@click.option('--project', '-p', 'project', required=True)
def export_pdf(src, project):
    click.echo('garjus! pdf')
    g = Garjus()
    g.pdf(src, project)


@cli.command('image03csv')
@click.option('--project', '-p', 'project', required=True)
@click.option(
    '--start', '-s', 'startdate', type=click.DateTime(formats=['%Y-%m-%d']))
@click.option(
    '--end', '-e', 'enddate', type=click.DateTime(formats=['%Y-%m-%d']))
@click.option('--site', multiple=True)
def image03csv(project, startdate, enddate, site):
    click.echo('garjus! image03csv')
    g = Garjus()
    g.image03csv(project, startdate, enddate, site)


@cli.command('retry')
@click.option('--project', '-p', 'project', required=True)
def retry(project):
    click.echo('garjus! retry')
    g = Garjus()
    g.retry(project)


@cli.command('image03download')
@click.argument('image03_csv', required=True)
@click.argument('download_dir', required=True)
@click.option('--project', '-p', 'project', required=True)
def image03download(project, image03_csv, download_dir):
    click.echo('garjus! image03download')
    g = Garjus()
    g.image03download(project, image03_csv, download_dir)


@cli.command('delete')
@click.option('--project', '-p', 'project', required=True)
@click.option('--type', '-t', 'proctype', required=True)
@click.option('--procstatus', '-s', 'procstatus', required=False)
@click.option('--qcstatus', '-q', 'qcstatus', required=False)
def delete(project, proctype, procstatus=None, qcstatus=None):
    click.echo('garjus! delete')
    g = Garjus()
    g.delete_proctype(
        project,
        proctype,
        procstatus=procstatus,
        qcstatus=qcstatus)


@cli.command('cleanup')
def cleanup():
    click.echo('garjus! cleanup')
    g = Garjus()
    g.delete_bad_tasks()


@cli.command('dashboard')
@click.option('--auth', 'auth_file', required=False)
@click.option('--login', required=False, is_flag=True)
@click.option('--demo', required=False, is_flag=True)
def dashboard(auth_file=None, login=False, demo=False):
    import webbrowser
    url = 'http://localhost:8050'

    if demo:
        from .dashboard.demo import app
    elif login:
        from .dashboard.login import app
    else:
        from .dashboard.index import app

        if auth_file:
            import yaml
            import dash_auth

            # Load user passwords to use dash's basic authentication
            with open(auth_file, 'rt') as file:
                _data = yaml.load(file, yaml.SafeLoader)
                dash_auth.BasicAuth(
                    app,
                    _data['VALID_USERNAME_PASSWORD_PAIRS']
                )

    # Open URL in a new tab, if a browser window is already open.
    webbrowser.open_new_tab(url)

    # start up a dashboard app
    app.run_server(host='0.0.0.0')


@cli.command('quicktest')
def quicktest():
    click.echo('garjus!')

    try:
        redcap_project = Garjus._default_redcap()
    except Exception:
        click.echo('could not connect to REDCap')
        return

    try:
        xnat_interface = Garjus._default_xnat()
    except Exception:
        click.echo('could not connect to XNAT')
        return

    try:
        g = Garjus(redcap_project, xnat_interface)
    except Exception:
        click.echo('something went wrong')
        return

    g.scans()
    click.echo('all good!')



"""Creates report PDF with zip."""
import logging
import io
from datetime import datetime
import math

import numpy as np
import pandas as pd
import plotly
import plotly.graph_objs as go
import plotly.subplots
from fpdf import FPDF
from fpdf.enums import XPos, YPos
from PIL import Image


logger = logging.getLogger('garjus.progress.report')


# These are used to set colors of graphs
RGB_DKBLUE = 'rgb(59,89,152)'
RGB_BLUE = 'rgb(66,133,244)'
RGB_GREEN = 'rgb(15,157,88)'
RGB_YELL = 'rgb(244,160,0)'
RGB_RED = 'rgb(219,68,55)'
RGB_PURP = 'rgb(160,106,255)'
RGB_GREY = 'rgb(200,200,200)'
RGB_PINK = 'rgb(255,182,193)'
RGB_LIME = 'rgb(17, 180, 101)'

# Give each status a color to display
QASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NQA': RGB_LIME,
    'NPUT': RGB_YELL,
    'FAIL': RGB_RED,
    'NONE': RGB_GREY,
    'JOBF': RGB_PINK,
    'JOBR': RGB_BLUE}

STATUS2RGB = dict(zip(
    ['WAITING', 'PENDING', 'RUNNING', 'COMPLETE', 'FAILED', 'UNKNOWN', 'JOBF'],
    [RGB_GREY, RGB_YELL, RGB_GREEN, RGB_BLUE, RGB_RED, RGB_PURP, RGB_PINK]))

# These are used to make progress reports
ASTATUS2COLOR = {
    'PASS': RGB_GREEN,
    'NPUT': RGB_YELL,
    'FAIL': RGB_RED,
    'NQA': RGB_LIME,
    'NONE': RGB_GREY,
    'COMPLETE': RGB_BLUE,
    'UNKNOWN': RGB_PURP}

HIDECOLS = [
    'assessor_label',
    'PROJECT',
    'SESSION',
    'SUBJECT',
    'AGE',
    'SEX',
    'DEPRESS',
    'TYPE',
    'SITE',
    'SESSTYPE',
    'DATE',
    'ASSR',
    'PROCSTATUS',
    'QCSTATUS',
    'INPUTS',
    'MODALITY',
    'XSITYPE',
    'PROCTYPE',
    'full_path',
    'case',
]


class MYPDF(FPDF):
    """Custom PDF."""

    def set_filename(self, filename):
        """Set the filename."""
        self.filename = filename

    def set_project(self, project, enable_monthly=False):
        """Set the project name."""
        self.project = project
        today = datetime.now().strftime("%Y-%m-%d")
        self.date = today

        if enable_monthly:
            self.title = f'{project} Monthly Report'
            self.subtitle = '{}'.format(datetime.now().strftime("%B %Y"))
        else:
            self.title = f'{project} Report'
            self.subtitle = '{}'.format(datetime.now().strftime("%B %d, %Y"))

    def set_date(self):
        today = datetime.now().strftime("%Y-%m-%d")
        self.date = today

    def footer(self):
        """Return the custom footer."""
        self.set_y(-0.35)
        self.set_x(0.5)

        # Write date, title, page number
        self.set_font('helvetica', size=10)
        self.set_text_color(100, 100, 100)
        self.set_draw_color(100, 100, 100)
        self.line(x1=0.2, y1=10.55, x2=8.3, y2=10.55)
        self.cell(w=1, text=self.date)
        self.cell(w=5, align='C', text=self.title)
        self.cell(w=2.5, align='C', text=str(self.page_no()))


def blank_letter():
    """Blank letter sized PDF."""
    p = MYPDF(orientation="P", unit='in', format='letter')
    p.set_top_margin(0.5)
    p.set_left_margin(0.5)
    p.set_right_margin(0.5)

    return p


def _draw_counts(pdf, subjects):
    # Counts of each type with sums
    project_list = sorted(subjects.PROJECT.unique())
    group_list = subjects.GROUP.unique()
    indent_width = max(2.5 - len(project_list) * 0.5, 0.3)

    pdf.set_fill_color(94, 156, 211)
    _txt = 'Participant Groups'

    # Draw heading
    pdf.set_font('helvetica', size=14)
    pdf.cell(w=7.5, h=0.5, text=_txt, align='C', border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Header Formatting
    pdf.cell(w=0.7)
    pdf.set_text_color(245, 245, 245)
    pdf.set_line_width(0.01)
    _kwargs = {'w': 0.9, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': True}

    # Column header for each project
    pdf.cell(indent_width)
    for cur_proj in project_list:
        _txt = cur_proj
        if len(_txt) > 6:
            pdf.set_font('helvetica', size=9)
        else:
            pdf.set_font('helvetica', size=12)

        pdf.cell(**_kwargs, text=_txt)

    # Got to next line
    pdf.ln()

    # Row formatting
    pdf.set_fill_color(255, 255, 255)
    pdf.set_text_color(0, 0, 0)
    _kwargs = {'w': 0.9, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_s = {'w': 0.7, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_t = {'w': 0.5, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}

    # Row for each group
    for cur_group in group_list:
        pdf.cell(w=indent_width)

        dfg = subjects[subjects.GROUP == cur_group]

        # Show the group
        _txt = cur_group
        if _txt == 'UNKNOWN':
            _txt = '?'

        if len(_txt) > 6:
            pdf.set_font('helvetica', size=12)
        else:
            pdf.set_font('helvetica', size=14)

        pdf.cell(**_kwargs_s, text=_txt)

        # Count each type for this group
        pdf.set_font('helvetica', size=14)
        for cur_proj in project_list:
            cur_count = str(len(dfg[dfg.PROJECT == cur_proj]))
            pdf.cell(**_kwargs, text=cur_count)

        if len(project_list) > 1:
            # Total for group
            cur_count = str(len(dfg))
            pdf.cell(**_kwargs_t, text=cur_count, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
        else:
            #pdf.cell(w=0, h=0, border=0, text='', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
            pdf.ln()

    pdf.set_font('helvetica', size=14)
    if len(group_list) > 1:
        # TOTALS row
        pdf.cell(w=indent_width)
        pdf.cell(w=0.7)
        for cur_proj in project_list:
            cur_count = str(len(subjects[subjects.PROJECT == cur_proj]))
            pdf.cell(**_kwargs, text=cur_count)

        # Grandtotal
        if len(project_list) > 1:
            pdf.cell(**_kwargs_t, text=str(len(subjects)))

    # End by going to next line
    pdf.ln()

    return pdf


def _draw_demog(pdf, subjects):
    # Counts of each type with sums
    project_list = sorted(subjects.PROJECT.unique())
    #sex_list = subjects.SEX.unique()
    sex_list = ['F']
    indent_width = max(2.5 - len(project_list) * 0.5, 0.3)

    pdf.set_fill_color(114, 172, 77)

    _txt = 'Participant Demographics'

    # Draw heading
    pdf.set_font('helvetica', size=14)
    pdf.cell(w=7.5, h=0.5, text=_txt, align='C', border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Header Formatting
    pdf.cell(w=0.5)
    pdf.set_text_color(245, 245, 245)
    pdf.set_line_width(0.01)
    _kwargs = {'w': 1.0, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': True}

    # Column header for each project
    pdf.cell(indent_width)
    for cur_proj in project_list:
        _txt = cur_proj
        if len(_txt) > 6:
            pdf.set_font('helvetica', size=10)

        pdf.cell(**_kwargs, text=_txt)

    # Got to next line
    pdf.ln()

    # Row formatting
    pdf.set_fill_color(255, 255, 255)
    pdf.set_text_color(0, 0, 0)
    _kwargs = {'w': 1.0, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_s = {'w': 0.5, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}
    _kwargs_t = {'w': 0.5, 'h': 0.5, 'border': 1, 'align': 'C', 'fill': False}


    # Row for each sex
    for cur_sex in sex_list:
        pdf.cell(w=indent_width)

        dfg = subjects[subjects.SEX == cur_sex]

        # Show this
        _txt = f'% {cur_sex}'

        if len(_txt) > 6:
            pdf.set_font('helvetica', size=12)
        else:
            pdf.set_font('helvetica', size=14)

        pdf.cell(**_kwargs_s, text=_txt)

        # Count each type for this
        pdf.set_font('helvetica', size=14)
        for cur_proj in project_list:
            proj_tot = len(subjects[subjects.PROJECT == cur_proj])
            proj_sex = len(dfg[dfg.PROJECT == cur_proj])
            cur_pct = str(int(proj_sex / proj_tot * 100 + 0.5))
            pdf.cell(**_kwargs, text=cur_pct)

        pdf.ln()

    pdf.cell(w=indent_width)
    pdf.cell(**_kwargs_s, text='Age')

    for cur_proj in project_list:
        cur_age = subjects[(subjects.PROJECT == cur_proj) & (subjects.AGE != '')].AGE.astype(float).mean()
        pdf.cell(**_kwargs, text=str(int(cur_age)))

    # End by going to next line
    pdf.ln()

    return pdf


def _draw_proc(pdf, stats):
    common_count = len(stats.dropna(axis=1).columns)
    if common_count == len(stats.columns):
        common_count = 8 # ASSR,PROJECT,SUBJECT,SESSION,SESSTYPE,SITE,DATE,PROCTYPE

    # Draw heading
    pdf.set_font('helvetica', size=14)
    _txt = 'Processing Types (see additional pages for details)\n' 
    pdf.cell(w=7.5, h=0.5, text=_txt, align='C', border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font('helvetica', size=12)

    # Column header
    pdf.cell(1.5)
    pdf.set_line_width(0.01)
    pdf.set_text_color(245, 245, 245)
    pdf.set_fill_color(232, 108, 32)
    pdf.cell(w=2.0, h=0.3, border=1, align='C', text='Type', fill=True)
    pdf.cell(w=1.0, h=0.3, border=1, align='C', text='Rows', fill=True)
    pdf.cell(w=1.0, h=0.3, border=1, align='C', text='Columns', fill=True)
    pdf.ln()

    pdf.set_fill_color(255, 255, 255)
    pdf.set_text_color(0, 0, 0)
    for proc in sorted(stats.PROCTYPE.unique()):
        pdf.cell(1.5)
        pstats = stats[stats.PROCTYPE == proc]
        _txt = proc
        pdf.cell(w=2.0, h=0.3, border=1, text=_txt)
        _txt = str(len(pstats))
        pdf.cell(w=1.0, h=0.3, border=1, align='C', text=_txt)
        _txt = str(len(pstats.dropna(axis=1).columns) - common_count)
        pdf.cell(w=1.0, h=0.3, border=1, align='R', text=_txt, new_x=XPos.LMARGIN, new_y=YPos.NEXT)


def _add_first_page(pdf, info):
    subjects = info['subjects']
    stats = info['stats']

    # Start the page with titles
    pdf.add_page()
    pdf.set_font('helvetica', size=16)
    pdf.cell(w=7.5, h=0.4, align='C', text=pdf.title, new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font('helvetica', size=12)
    pdf.cell(w=7.5, h=0.4, align='C', text=pdf.subtitle, border='B', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.ln(0.1)

    pdf.set_font('helvetica', size=12)

    # Show settings for proctypes, sesstypes, projects, etc.
    _text = 'Subject data in subjects.csv, processing data in [TYPE].csv, e.g. FS7_v1.csv\n'
    _text += f'XNAT: {info.get("xnat")}\n'
    _text +=f'REDCap:{info.get("redcap")}\n'
    _text += f'Session Types: '
    _text += ','.join(sorted(list(stats.SESSTYPE.unique())))
    pdf.multi_cell(0, 0.25, text=_text, new_x=XPos.LMARGIN, new_y=YPos.NEXT, align='L')

    # Show subject counts by project
    _draw_counts(pdf, subjects)
    pdf.ln(0.2)

    # Show demographics table
    if not subjects.isna().any().any():
        _draw_demog(pdf, subjects)
        pdf.ln(0.2)
    else:
        logger.info('missing Demographics')

    # Show processing types table
    _draw_proc(pdf, stats)

    return pdf

def _add_stats_fmriqa(pdf, stats, info):
    # get the stats data by scan type using inputs field to map to scan

    scans = info['scans']

    assessors = info['assessors'].copy()
    assessors = assessors[assessors.PROCTYPE == 'fmriqa_v4']

    # Extract scan fmri value into a column
    assessors['scan_fmri'] = assessors.apply(
        lambda row: row['INPUTS'].get('scan_fmri'), axis=1)

    df = pd.merge(
        assessors[['ASSR', 'scan_fmri']],
        scans[['full_path', 'SCANTYPE']],
        how='inner',
        left_on='scan_fmri',
        right_on='full_path')

    df = pd.merge(
        stats,
        df[['ASSR', 'SCANTYPE']],
        how='inner',
        left_on='ASSR',
        right_on='ASSR')

    for t in sorted(df.SCANTYPE.unique()):
        if t in ['fMRI_rest', 'fMRI_REST_FSA']:
            continue

        pdf.set_font('helvetica', size=10)
        _add_stats(pdf, df[df.SCANTYPE == t], plot_title=f'Scan Type: {t}')


def _add_stats(pdf, stats, plot_title=None):

    # this returns a PIL Image object
    image = plot_stats(stats, plot_title)
    tot_width, tot_height = image.size

    # Split horizontal image into chunks of width to fit on
    # letter-sized page with crop((left, top, right, bottom))
    chunk_h = 500
    chunk_w = 998
    rows_per_page = 3  # 3 rows per page
    page_count = math.ceil(tot_width / (rows_per_page * chunk_w))

    for p in range(page_count):
        for c in range(rows_per_page):
            # Calculate the starting x for this chunk
            chunk_x = (c * chunk_w) + (p * chunk_w * rows_per_page)
            if chunk_x > tot_width - 10:
                # out of bounds
                continue

            # Get the image from the cropped section
            _img = image.crop((chunk_x, 0, chunk_x + chunk_w, chunk_h))

            # Draw the image on the PDF
            pdf.image(_img, x=0.75, h=3.1)

    return pdf


def _add_covar(pdf, covar, plot_title=None):
    # this returns a PIL Image object
    image = plot_covar(covar, plot_title)
    tot_width, tot_height = image.size

    # Split horizontal image into chunks of width to fit on
    # letter-sized page with crop((left, top, right, bottom))
    chunk_h = 500
    chunk_w = 998
    rows_per_page = 3  # 3 rows per page
    page_count = math.ceil(tot_width / (rows_per_page * chunk_w))

    for p in range(page_count):
        for c in range(rows_per_page):
            # Calculate the starting x for this chunk
            chunk_x = (c * chunk_w) + (p * chunk_w * rows_per_page)
            if chunk_x > tot_width - 10:
                # out of bounds
                continue

            # Get the image from the cropped section
            _img = image.crop((chunk_x, 0, chunk_x + chunk_w, chunk_h))

            # Draw the image on the PDF
            pdf.image(_img, x=0.75, h=3.1)

    return pdf


def _plottable(var):
    try:
        _ = var.astype(str).replace('', np.nan).dropna().str.strip('%').astype(float)
        return True
    except Exception:
        return False


def plot_stats(df, plot_title=None):
    """Plot stats, one boxlplot per var."""
    box_width = 250
    min_box_count = 4

    logger.debug('plot_stats:{}'.format(len(df)))

    # Check for empty data
    if len(df) == 0:
        logger.debug('empty data, using empty figure')
        fig = go.Figure()
        _png = fig.to_image(format="png")
        image = Image.open(io.BytesIO(_png))
        return image

    # Filter var list to only include those that have data
    var_list = [x for x in df.columns if not pd.isnull(df[x]).all()]

    # Filter var list to only stats variables
    var_list = [x for x in var_list if x not in HIDECOLS]

    # Filter var list to only stats can be plotted as float
    var_list = [x for x in var_list if _plottable(df[x])]

    # skip vars
    var_list = [x for x in var_list if not x.endswith('pathlength')]

    # Determine how many boxplots we're making, depends on how many vars, use
    # minimum so graph doesn't get too small
    box_count = len(var_list)
    site_count = len(df['SITE'].unique())

    if box_count < 3:
        box_width = 500
        min_box_count = 2
    elif box_count < 6:
        box_width = 333
        min_box_count = 3

    if site_count > 3:
        box_width = 500
        min_box_count = 2

    if box_count < min_box_count:
        box_count = min_box_count

    graph_width = box_width * box_count

    # Horizontal spacing cannot be greater than (1 / (cols - 1))
    hspacing = 1 / (box_count * 4)

    logger.debug(f'{box_count}, {min_box_count}, {graph_width}, {hspacing}')

    # Make the figure with 1 row and a column for each var we are plotting
    var_titles = [x[:22] for x in var_list]
    fig = plotly.subplots.make_subplots(
        rows=1,
        cols=box_count,
        horizontal_spacing=hspacing,
        subplot_titles=var_titles)

    # Add box plot for each variable
    for i, var in enumerate(var_list):

        # Create boxplot for this var and add to figure
        logger.debug('plotting var:{}'.format(var))

        _row = 1
        _col = i + 1

        fig.append_trace(
            go.Box(
                y=df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float),
                x=df['SITE'],
                boxpoints='all',
                text=df['ASSR'],
                boxmean=True,
            ),
            _row,
            _col)

        # Plot horizontal line at median
        _median = df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float).median()
        fig.add_trace(
            go.Scatter(
                x=df['SITE'],
                y=[_median] * len(df),
                mode='lines',
                name='',
                fillcolor='red',
                line_color='grey'),
            _row, _col)

        fig.update_yaxes(autorange=True)

        if var.startswith('con_') or var.startswith('inc_'):
            logger.debug('setting beta range:{}'.format(var))
            _yaxis = 'yaxis{}'.format(i + 1)
            fig['layout'][_yaxis].update(range=[-1, 1], autorange=False)
        else:
            pass

    # Move the subtitles to bottom instead of top of each subplot
    if len(df['SITE'].unique()) < 4:
        for i in range(len(fig.layout.annotations)):
            fig.layout.annotations[i].update(y=-.15)

    if plot_title:
        fig.update_layout(
            title={'text': plot_title, 'x': 0.5, 'xanchor': 'center'})

    # Customize figure to hide legend and fit the graph
    fig.update_layout(
        showlegend=False,
        autosize=False,
        width=graph_width,
        margin=dict(l=20, r=40, t=40, b=80, pad=0))

    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))
    return image


def plot_covar(df, plot_title=None):
    """Plot one boxlplot per var."""
    box_width = 250
    min_box_count = 4

    logger.debug(f'plot_covar:{len(df)}')

    # Check for empty data
    if len(df) == 0:
        logger.debug('empty data, using empty figure')
        fig = go.Figure()
        _png = fig.to_image(format="png")
        image = Image.open(io.BytesIO(_png))
        return image

    # Filter var list to only include those that have data
    var_list = [x for x in df.columns if not pd.isnull(df[x]).all()]

    # Filter var list to no hidden variables
    var_list = [x for x in var_list if x not in HIDECOLS]

    # Filter var list to only variables that can be plotted as float
    var_list = [x for x in var_list if _plottable(df[x])]

    # Determine how many boxplots we're making, depends on how many vars, use
    # minimum so graph doesn't get too small
    box_count = len(var_list)
    site_count = len(df['SITE'].unique())

    if box_count < 3:
        box_width = 500
        min_box_count = 2
    elif box_count < 6:
        box_width = 333
        min_box_count = 3

    if site_count > 3:
        box_width = 500
        min_box_count = 2

    if box_count < min_box_count:
        box_count = min_box_count

    graph_width = box_width * box_count

    # Horizontal spacing cannot be greater than (1 / (cols - 1))
    hspacing = 1 / (box_count * 4)

    logger.debug(f'{box_count}, {min_box_count}, {graph_width}, {hspacing}')

    # Make the figure with 1 row and a column for each var we are plotting
    var_titles = [x[:22] for x in var_list]
    fig = plotly.subplots.make_subplots(
        rows=1,
        cols=box_count,
        horizontal_spacing=hspacing,
        subplot_titles=var_titles)

    # Add box plot for each variable
    for i, var in enumerate(var_list):
        # Create boxplot for this var and add to figure
        logger.debug('plotting var:{}'.format(var))

        _row = 1
        _col = i + 1

        fig.append_trace(
            go.Box(
                y=df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float),
                x=df['SITE'],
                boxpoints='all',
                boxmean=True,
            ),
            _row,
            _col)

        # Plot horizontal line at median
        _median = df[var].astype(str).replace('', np.nan).dropna().str.strip('%').astype(float).median()
        fig.add_trace(
            go.Scatter(
                x=df['SITE'],
                y=[_median] * len(df),
                mode='lines',
                name='',
                fillcolor='red',
                line_color='grey'),
            _row, _col)

        fig.update_yaxes(autorange=True)

        if var.startswith('con_') or var.startswith('inc_'):
            logger.debug('setting beta range:{}'.format(var))
            _yaxis = 'yaxis{}'.format(i + 1)
            fig['layout'][_yaxis].update(range=[-1, 1], autorange=False)
        else:
            pass

    # Move the subtitles to bottom instead of top of each subplot
    if len(df['SITE'].unique()) < 4:
        for i in range(len(fig.layout.annotations)):
            fig.layout.annotations[i].update(y=-.15)

    if plot_title:
        fig.update_layout(
            title={'text': plot_title, 'x': 0.5, 'xanchor': 'center'})

    # Customize figure to hide legend and fit the graph
    fig.update_layout(
        showlegend=False,
        autosize=False,
        width=graph_width,
        margin=dict(l=20, r=40, t=40, b=80, pad=0))

    _png = fig.to_image(format="png")
    image = Image.open(io.BytesIO(_png))
    return image


def _add_stats_pages(pdf, info):
    proclib = info['proclib']
    stats = info['stats']
    stattypes = info['stattypes']

    for proctype in stattypes:
        # Limit the data to this proctype
        stat_data = stats[stats.PROCTYPE == proctype]

        if stat_data.empty:
            logger.debug(f'no stats for proctype:{proctype}')
            continue

        logger.debug(f'add stats page:{proctype}')

        # Get descriptions for this processing type
        proc_info = proclib.get(proctype, {})

        # use proclib to filter stats variable names
        _subset = proc_info.get('stats_subset', None)
        if _subset:
            stat_data = stat_data[_subset + ['SITE', 'ASSR']]
       
        # Now make the page
        pdf.add_page()
        pdf.set_font('helvetica', size=14)
        pdf.cell(text=proctype, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        if proctype == 'fmriqa_v4':
            # stats by scan type using inputs field to map to scan
            _add_stats_fmriqa(pdf, stat_data, info)
        else:
            _add_stats(pdf, stat_data)

        # Build the description
        _text = proc_info.get('short_descrip', '') + '\n'
        _text += 'Inputs: ' + proc_info.get('inputs_descrip', '') + '\n'

        # Append stats descriptions
        for s, t in info['statlib'].get(proctype, {}).items():
            _text += f'{s}: {t}\n'

        # Show the descriptions
        pdf.set_font('helvetica', size=12)
        pdf.multi_cell(0, 0.25, _text, border='LBTR', align="L", new_x=XPos.RIGHT, new_y=YPos.NEXT)

        _url = proc_info.get('procurl', '')
        if _url:
            pdf.set_font('helvetica', size=10)
            pdf.cell(text=_url, link=_url)


def _add_covar_pages(pdf, info):
    covariates = info['covariates']
    proclib = info['proclib']

    for k, values in covariates.items():
        logger.debug(f'add covar page:{k}')

        covar_data = values

        if covar_data.empty:
            logger.debug(f'no stats for proctype:{k}')
            continue

        # Get descriptions for this processing type
        proc_info = proclib.get(k, {})

        # use proclib to filter stats variable names
        _subset = proc_info.get('stats_subset', None)
        if _subset:
            covar_data = covar_data[_subset + ['SITE']]

        # Now make the page
        pdf.add_page()
        pdf.set_font('helvetica', size=14)
        _text = f'{k} (n={len(values)})'
        pdf.cell(text=_text, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

        _add_covar(pdf, covar_data)

        # Build the description
        _text = proc_info.get('short_descrip', '') + '\n'

        # Append stats descriptions
        for s, t in info['statlib'].get(k, {}).items():
            _text += f'{s}: {t}\n'

        # Show the descriptions
        pdf.set_font('helvetica', size=12)
        pdf.multi_cell(0, 0.25, _text, border='LBTR', align="L", new_x=XPos.RIGHT, new_y=YPos.NEXT)

        _url = proc_info.get('procurl', '')
        if _url:
            pdf.set_font('helvetica', size=10)
            pdf.cell(text=_url, link=_url)


def make_export_report(filename, garjus, subjects, stats, covar):
    # Initialize a new PDF letter size and shaped
    pdf = blank_letter()
    pdf.set_filename(filename)
    pdf.set_date()
    pdf.title = f'Data Export Report'
    pdf.subtitle = '{}'.format(datetime.now().strftime("%B %d, %Y"))
    info = {}

    p2p = {
        'TAYLOR_CAARE': 'CAARE',
        'NewhouseMDDHx': 'MDDHx'
    }
    subjects['PROJECT'] = subjects.PROJECT.replace(p2p)
    stats['PROJECT'] = stats.PROJECT.replace(p2p)
    subjects['SEX'] = subjects.SEX.fillna('')

    info['xnat'] = garjus.xnat_host()
    info['redcap'] = 'https://redcap.vumc.org/redcap_v14.8.1/index.php?pid=156730'
    info['proclib'] = garjus.processing_library()
    info['subjects'] = subjects
    info['stats'] = stats
    info['stattypes'] = stats.PROCTYPE.unique()
    info['statlib'] = garjus.stats_library()
    info['covariates'] = covar

    logger.debug('adding first page')
    _add_first_page(pdf, info)

    logger.debug('adding stats pages')
    _add_stats_pages(pdf, info)

    logger.debug('adding covariate pages')
    _add_covar_pages(pdf, info)

    # Save to file
    print(f'saving pdf:{pdf.filename}')
    logger.debug('saving PDF to file:{}'.format(pdf.filename))
    try:
        pdf.output(pdf.filename)
    except Exception as err:
        logger.error('error while saving PDF:{}:{}'.format(pdf.filename, err))
        return False

    return True



import glob
import re
import logging

from ...utils_redcap import upload_file, field2events


logger = logging.getLogger('garjus.automations.etl_nihexaminer.files2redcap')


# Finds NIH examiner summary files and uploads to REDCap.


def process_project(
    project,
    limbo_dir,
    event2sess
):
    results = []
    def_field = project.def_field
    id2subj = {}
    flank_field = 'flanker_file'
    nback_field = 'nback_upload'
    shift_field = 'set_shifting_file'
    cpt_field = 'cpt_upload'
    done_field = 'flanker_score'

    if 'flanker_summfile' in project.field_names:
        # Alternate file field names used in some old REDCap projects
        flank_field = 'flanker_summfile'
        nback_field = 'nback_summfile'
        shift_field = 'set_shifting_summfile'
        cpt_field = 'cpt_summfile'

    # Set the names of the fields we want to query
    fields = [def_field, flank_field, nback_field, shift_field, cpt_field, done_field]

    # Determine events
    events = field2events(project, cpt_field)

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event = r['redcap_event_name']
        sess_num = '1'

        if event2sess is not None:
            sess_num = event2sess[event]

        try:
            subj = id2subj[record_id]
            logger.debug(f'{subj}:{event} using secondary id:{record_id}')
        except KeyError as err:
            logger.debug(f'record without subject number:{err}')
            continue

        if not subj:
            logger.warn(f'blank subject number:{record_id}')
            continue

        # Find files for this record
        logger.debug(f'looking for files:{subj}:{event}')

        # Get just the numeric portion of the subject number
        subj_num = re.sub(r'[^0-9]', '', subj)

        # Find files for this subject
        for cur_field, cur_name in zip(
            [cpt_field, nback_field, flank_field, shift_field], 
            ['CPT', 'NBack', 'Flanker', 'SetShifting']):

            # Check existing
            if r[cur_field]:
                logger.debug(f'already uploaded:{subj}:{event}:{cur_field}')
                continue

            cur_glob = f'{limbo_dir}/{subj_num}/{cur_name}_Summary_{subj_num}_{sess_num}_*.csv'
            file_list = sorted(glob.glob(cur_glob))
            file_count = len(file_list)
            if file_count <= 0:
                logger.debug(f'no file found:{subj}:{event}:{cur_glob}')
                continue
            elif file_count > 1:
                logger.debug(f'too many files:{subj}:{event}:{cur_glob}')
                continue

            #Upload file to redcap
            cur_file = file_list[0]
            logger.debug(f'uploading file:{cur_file}')

            try:
               result = upload_file(
                   project,
                   record_id,
                   cur_field,
                   cur_file,
                   event_id=event)

               logger.debug(f'uploaded:{subj}:{event}:{cur_file}')
            except (ValueError) as err:
               logger.error(f'error uploading:{cur_file}:{err}')

            if not result:
               logger.error(f'upload failed:{subj}:{event}')
               continue

            logger.debug(f'uploaded:{subj}:{event}')
            results.append({
               'result': 'COMPLETE',
               'description': 'etl_nihexaminer.file2redcap',
               'category': 'etl_nihexaminer.file2redcap',
               'subject': subj,
               'session': '',
               'scan': '',
               'event': event,
               'field': cur_field})

    return results



import logging
from datetime import datetime

import pandas as pd


logger = logging.getLogger('garjus.automations.etl_arcapp.dailyweekly')


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False


# Load ARC tests and calculate summary measures
def extract_arc(records):
    data = {}

    # Load into dataframe
    df = pd.DataFrame(records)

    # Count complete sessions
    data['arc_warcsesscomp'] = len(df[df['arc_finishedsession'] == '1'])

    # Convert format to float so we can math
    df['arc_symbolsrt'] = pd.to_numeric(
        df['arc_symbolsrt'], errors='coerce')
    df['arc_symbols_accuracy'] = pd.to_numeric(
        df['arc_symbols_accuracy'], errors='coerce')
    df['arc_pricesrt'] = pd.to_numeric(
        df['arc_pricesrt'], errors='coerce')
    df['arc_prices_accuracy'] = pd.to_numeric(
        df['arc_prices_accuracy'], errors='coerce')
    df['arc_grided'] = pd.to_numeric(
        df['arc_grided'], errors='coerce')

    # Calculate range for whole week
    data[f'arc_wrangesymbolsrt'] = \
        df.arc_symbolsrt.max() - \
        df.arc_symbolsrt.min()
    data[f'arc_wrangesymbolsacc'] = \
        df.arc_symbols_accuracy.max() - \
        df.arc_symbols_accuracy.min()
    data[f'arc_wrangepricesrt'] = \
        df.arc_pricesrt.max() - \
        df.arc_pricesrt.min()
    data[f'arc_wrangepricesacc'] = \
        df.arc_prices_accuracy.max() - \
        df.arc_prices_accuracy.min()
    data[f'arc_wrangegrided'] = \
        df.arc_grided.max() - \
        df.arc_grided.min()

    # Calculate means for whole week
    data['arc_wmeansymbolsrt'] = df.arc_symbolsrt.mean()
    data['arc_wmeansymbolsacc'] = df.arc_symbols_accuracy.mean()
    data['arc_wmeanpricesrt'] = df.arc_pricesrt.mean()
    data['arc_wmeanpricesacc'] = df.arc_prices_accuracy.mean()
    data['arc_wmeangrided'] = df.arc_grided.mean()

    # Calculate SD for whole week
    data['arc_wsdsymbolsrt'] = df.arc_symbolsrt.std()
    data['arc_wsdsymbolsacc'] = df.arc_symbols_accuracy.std()
    data['arc_wsdpricesrt'] = df.arc_pricesrt.std()
    data['arc_wsdpricesacc'] = df.arc_prices_accuracy.std()
    data['arc_wsdgrided'] = df.arc_grided.std()

    # Calculate CV for whole week
    data[f'arc_wcovsymbolsrt'] = \
        df.arc_symbolsrt.std() / \
        df.arc_symbolsrt.mean()
    data[f'arc_wcovsymbolsacc'] = \
        df.arc_symbols_accuracy.std() / \
        df.arc_symbols_accuracy.mean()
    data[f'arc_wcovpricesrt'] = \
        df.arc_pricesrt.std() / \
        df.arc_pricesrt.mean()
    data[f'arc_wcovpricesacc'] = \
        df.arc_prices_accuracy.std() / \
        df.arc_prices_accuracy.mean()
    data[f'arc_wcovgrided'] = \
        df.arc_grided.std() / \
        df.arc_grided.mean()

    # Loop days 1-7, calculate for each day
    for d in range(1, 8):
        dfd = df[df['arc_day_index'] == str(d)]

        # Count complete sessions
        data[f'arc_d{d}arcsesscomp'] = \
            len(dfd[dfd['arc_finishedsession'] == '1'])

        # Calculate ranges
        data[f'arc_d{d}rangesymbolsrt'] = \
            dfd.arc_symbolsrt.max() - \
            dfd.arc_symbolsrt.min()
        data[f'arc_d{d}rangesymbolsacc'] = \
            dfd.arc_symbols_accuracy.max() - \
            dfd.arc_symbols_accuracy.min()
        data[f'arc_d{d}rangepricesrt'] = \
            dfd.arc_pricesrt.max() - \
            dfd.arc_pricesrt.min()
        data[f'arc_d{d}rangepricesacc'] = \
            dfd.arc_prices_accuracy.max() - \
            dfd.arc_prices_accuracy.min()
        data[f'arc_d{d}rangegrided'] = \
            dfd.arc_grided.max() - \
            dfd.arc_grided.min()

        # Calculate means
        data[f'arc_d{d}meansymbolsrt'] = dfd.arc_symbolsrt.mean()
        data[f'arc_d{d}meansymbolsacc'] = dfd.arc_symbols_accuracy.mean()
        data[f'arc_d{d}meanpricesrt'] = dfd.arc_pricesrt.mean()
        data[f'arc_d{d}meanpricesacc'] = dfd.arc_prices_accuracy.mean()
        data[f'arc_d{d}meangrided'] = dfd.arc_grided.mean()

        # Calculate SD
        data[f'arc_d{d}sdsymbolsrt'] = dfd.arc_symbolsrt.std()
        data[f'arc_d{d}sdsymbolsacc'] = dfd.arc_symbols_accuracy.std()
        data[f'arc_d{d}sdpricesrt'] = dfd.arc_pricesrt.std()
        data[f'arc_d{d}sdpricesacc'] = dfd.arc_prices_accuracy.std()
        data[f'arc_d{d}sdgrided'] = dfd.arc_grided.std()

        # Calculate CV
        data[f'arc_d{d}covsymbolsrt'] = \
            dfd.arc_symbolsrt.std() / \
            dfd.arc_symbolsrt.mean()
        data[f'arc_d{d}covsymbolsacc'] = \
            dfd.arc_symbols_accuracy.std() / \
            dfd.arc_symbols_accuracy.mean()
        data[f'arc_d{d}covpricesrt'] = \
            dfd.arc_pricesrt.std() / \
            dfd.arc_pricesrt.mean()
        data[f'arc_d{d}covpricesacc'] = \
            dfd.arc_prices_accuracy.std() / \
            dfd.arc_prices_accuracy.mean()
        data[f'arc_d{d}covgrided'] = \
            dfd.arc_grided.std() / \
            dfd.arc_grided.mean()

    # Return etl data as strings
    for k, v in data.items():
        data[k] = str(v)

    return data


def process_project(project):
    results = []
    def_field = project.def_field
    fields = [
        def_field,
        'arc_finishedsession',
        'arc_response_date',
        'arc_warcsesscomp'
    ]

    subj2id = {}
    subjects = []

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
        subjects = list(set([x[sec_field] for x in rec if x[sec_field]]))
        subjects = sorted(subjects)
    else:
        rec = project.export_records(fields=[def_field])
        subj2id = {x[def_field]: x[def_field] for x in rec if x[def_field]}
        subjects = sorted(list(set([x[def_field] for x in rec])))

    # Get records
    all_records = project.export_records(fields=fields, forms=['arc_data'])

    # Get the arcdata repeating records
    arc_records = [x for x in all_records if x['redcap_repeat_instance'] and x['redcap_repeat_instrument'] == 'arc_data']

    # Process each subject
    for subj in subjects:
        subj_id = subj2id[subj]
        subj_events = sorted(list(set([x['redcap_event_name'] for x in all_records if x[def_field] == subj_id])))
        subj_arc = [x for x in arc_records if x[def_field] == subj_id]

        # Iterate subject events
        for event_id in subj_events:

            # Find existing
            found = [x for x in all_records if (x[def_field] == subj_id) and (x['redcap_event_name'] == event_id) and (x.get('arc_warcsesscomp', False))]
            if len(found) > 0:
                logger.debug(f'found existing:{subj}:{event_id}')
                continue

            # Get repeat records
            repeats = [x for x in subj_arc if (x['redcap_event_name'] == event_id)]
            if len(repeats) == 0:
                # no repeats to count
                logger.debug(f'no repeats found:{subj}:{event_id}')
                continue

            # Check date of tests, if less than week since starting, skip
            dates = [x['arc_response_date'] for x in repeats if x['arc_response_date']]
            if len(dates) > 0:
                first_date = sorted(dates)[0]
                if (datetime.today() - datetime.strptime(first_date, '%Y-%m-%d')).days < 7:
                    logger.debug(f'SKIP:{subj}:{event_id}:{first_date}')
                    continue

            finished = [x for x in subj_arc if (x['redcap_event_name'] == event_id) and (x.get('arc_finishedsession', '0') == '1')]
            count_finished = len(finished)

            logger.debug(f'{subj}:{event_id}:{count_finished=}')

            data = {
                def_field: subj_id,
                'redcap_event_name': event_id,
                'arc_dailyweekly_complete': '2',
            }

            if count_finished > 0:
                # Calculate daily and weekly
                data.update(extract_arc(finished))
            else:
                logger.debug(f'no finished records found:{subj}:{event_id}')
                data.update({'arc_warcsesscomp': '0'})

            result = _load(project, [data])
            if result is True:
                results.append({
                    'result': 'COMPLETE',
                    'description': 'arc_dailyweekly',
                    'category': 'arc_dailyweekly',
                    'subject': subj_id,
                    'event': event_id
                })

    return results



"""Fitbit data extraction."""
import logging
import os
import tempfile

import pandas as pd


logger = logging.getLogger('garjus.automations.etl_fitbit')


def process(data_file):
    """Process file and return subset of data."""
    data = {}

    # Extract data from file
    df = _extract(data_file)

    # Get days with at least 50% worn
    df = df[df['Percentage Worn'].astype(float) >= 50.0]

    # Count the days worn
    data['fitbit_daysworn'] = len(df)

    return data


def _extract(filename):
    """ Extract data from file that has a header row and one data row"""
    try:
        df = pd.read_csv(filename, dtype=str)
    except Exception:
        df = pd.read_excel(filename, dtype=str)

    try:
        df = df.dropna(subset=['Percentage Worn'])
    except Exception as err:
        logger.error(f'failed to extract:{err}')
        return []

    # Fill nan with blanks
    df = df.fillna('')

    return df.to_records()


if __name__ == "__main__":
    import pprint

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s:%(module)s:%(message)s',
        level=logging.DEBUG,
        datefmt='%Y-%m-%d %H:%M:%S')

    _dir = os.path.expanduser('~/Downloads')
    test_file = f'{_dir}/TimeFBWorn.csv'

    data = process(test_file)
    pprint.pprint(data)



import logging
import os

import pandas as pd

from .. import utils
from ....garjus import Garjus


logger = logging.getLogger('dashboard.queue.data')


# This is where we save our cache of the data
def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/queuedata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def get_data(proj_filter, hidedone=True):
    g = Garjus()

    if not g.rcq_enabled():
        logger.debug('rcq not enabled, no data')
        return pd.DataFrame(columns=g.column_names('tasks'))

    df = g.tasks(hidedone=hidedone)

    df = df[df.STATUS != 'NEED_INPUTS']

    df.reset_index(inplace=True)
    df['LABEL'] = df['ASSESSOR']

    df = df.apply(_get_proctype, axis=1)

    return df


def _get_proctype(row):
    try:
        if row['YAMLUPLOAD']:
            tmp = os.path.basename(row['YAMLUPLOAD'])
        else:
            tmp = os.path.basename(row['YAMLFILE'])

        # Get just the filename without the directory path
        # Split on periods and grab the 4th value from right,
        # thus allowing periods in the main processor name
        row['PROCTYPE'] = tmp.rsplit('.')[-4]
        row['PROCESSOR'] = tmp
    except (KeyError, IndexError):
        row['PROCTYPE'] = ''
        row['PROCESSOR'] = ''

    return row


def run_refresh(hidedone=True):
    filename = get_filename()
    proj_filter = []

    df = get_data(proj_filter, hidedone=hidedone)

    utils.save_data(df, filename)

    return df


def load_data(refresh=False, hidedone=True, maxmins=5):
    filename = get_filename()

    if not os.path.exists(filename):
        refresh = True
    elif utils.file_age(filename) > maxmins:
        logger.info(f'refreshing, file age limit reached:{maxmins} minutes')
        refresh = True

    if refresh:
        df = run_refresh(hidedone=hidedone)
    else:
        df = utils.read_data(filename)

    logger.info('reading data from file:{}'.format(filename))
    return df


def filter_data(df, proj, proc, user):
    # Filter by project
    if proj:
        logger.debug(f'filtering by project:{proj}')
        df = df[df['PROJECT'].isin(proj)]

    # Filter by proc
    if proc:
        logger.debug(f'filtering by proc:{proc}')
        df = df[(df['PROCTYPE'].isin(proc))]

    # Filter by user
    if user:
        logger.debug(f'filtering by user:{user}')
        df = df[(df['USER'].isin(user))]

    return df



import logging
from datetime import datetime


logger = logging.getLogger('garjus.automations.etl_arcdata.summary')


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False


def process(project):
    results = []
    def_field = project.def_field
    fields = [def_field, 'arcapp_numcomplete', 'arc_finishedsession', 'arc_response_date']
    subj2id = {}
    subjects = []

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
        subjects = sorted(list(set([x[sec_field] for x in rec if x[sec_field]])))
    else:
        rec = project.export_records(fields=[def_field])
        subj2id = {x[def_field]: x[def_field] for x in rec if x[def_field]}
        subjects = sorted(list(set([x[def_field] for x in rec])))

    # Get all records, repeats and non-repeats
    all_records = project.export_records(fields=fields)

    # Get the arcdata repeating records
    arc_records = [x for x in all_records if x['redcap_repeat_instance']]

    # Process each subject
    for subj in subjects:
        subj_id = subj2id[subj]
        subj_events = list(set([x['redcap_event_name'] for x in all_records if x[def_field] == subj_id]))
        subj_arc = [x for x in arc_records if x[def_field] == subj_id]

        # Iterate subject events
        for event_id in subj_events:

            # Find existing numcomplete
            numcomplete = [x for x in all_records if (x[def_field] == subj_id) and (x['redcap_event_name'] == event_id) and (x.get('arcapp_numcomplete', False))]
            if len(numcomplete) > 0:
                # numcomplete already set
                continue

            # Get repeat records
            repeats = [x for x in subj_arc if (x['redcap_event_name'] == event_id) and (x.get('arc_response_date', False))]
            if len(repeats) == 0:
                # no repeats to count
                continue

            # Check date of tests, if less than week since starting, skip
            first_date = sorted([x['arc_response_date'] for x in repeats])[0]
            if (datetime.today() - datetime.strptime(first_date, '%Y-%m-%d')).days <  7:
                logger.debug(f'SKIP:{subj}:{event_id}:{first_date}')
                continue

            finished = [x for x in subj_arc if (x['redcap_event_name'] == event_id) and (x.get('arc_finishedsession', False) == '1')]
            count_finished = len(finished)

            logger.debug(f'setting numcomplete:{subj}:{event_id}:{first_date}:{count_finished}')

            # set numcomplete equal to count_finished for record/event
            data = {
                def_field: subj_id,
                'redcap_event_name': event_id,
                'arcapp_numcomplete': str(count_finished),
                'arc_app_complete': '2',
            }
            logger.debug(f'loading numcomplete:{subj_id}:{event_id}')

            _load(project, [data])

            results.append({
                'result': 'COMPLETE',
                'description': 'arc_summary',
                'category': 'arc_summary',
                'subject': subj_id,
                'event': event_id,
                'field': 'arcapp_numcomplete'})

    return results



"""dash index page."""
import logging

from .app import app
from . import content


logger = logging.getLogger('garjus.dashboard')


app.layout = content.get_content()

# For gunicorn to work correctly
server = app.server

# Allow external css
app.css.config.serve_locally = False

# Set the title to appear on web pages
app.title = 'dashboard'

if __name__ == '__main__':
    app.run_server(host='0.0.0.0')



"""Extract Context Data from old XML files"""
import json
import tempfile
import logging

import xml.etree.ElementTree as etree

from ...utils_redcap import field2events, download_file

# These files were saved to phones and an xml file exported for each event,
# Baseline, Month8, etc. The new version does not save these files, instead
# we receive a file per testing session, so file per row in the
# repeating instrument arcdata.


logger = logging.getLogger('garjus.automations.etl_arcapp.extract_context_xml')

raw_field = 'arcapp_xmlfile'

tab_field = 'arcapp_csvfile'


def process_project(project):
    results = []
    events = field2events(project, raw_field)
    def_field = project.def_field
    fields = [def_field, raw_field, tab_field]
    id2subj = {}

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if not sec_field:
        logger.error('secondary enabled, but no secondary field found')
        return

    rec = project.export_records(fields=[def_field, sec_field])
    id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event = r['redcap_event_name']

        try:
            subj = id2subj[record_id]
        except KeyError as err:
            print('record without subject number:', err)
            continue

        # Check for no raw file
        if not r[raw_field]:
            logger.debug('{}:{}:{}'.format(subj, event, 'no xml file'))
            continue

        if not r[tab_field]:
            logger.debug('{}:{}:{}'.format(subj, event, 'not converted'))
            continue

        if r[tab_field] in ['MISSING_DATA.txt', 'CONVERT_FAILED.txt']:
            logger.debug('{}:{}:{}'.format(subj, event, 'convert failed'))
            continue

        logger.info(f'extracting:{subj}:{event}')
        result = _process(project, record_id, event)

        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_arcapp.extract_context_xml',
                'description': 'etl_arcapp.extract_context_xml',
                'subject': subj,
                'event': event,
                'field': raw_field})

    return results


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False


def _process(project, record_id, event_id):
    logger.debug(f'etl_arcapp:{record_id}:{event_id}')

    # check for existing records in the repeating instruments for this event
    rec = project.export_records(
        records=[record_id],
        events=[event_id],
        forms=['arc_data'],
        fields=[project.def_field]
    )

    rec = [x for x in rec if x['redcap_repeat_instrument'] == 'arc_data']

    if len(rec) > 0:
        logger.debug(f'found existing records:{record_id}:{event_id}')
        count = len([x['arc_context1'] for x in rec if x['arc_context1']])
        symbol_count = len(
            [x['arc_symbolsrt'] for x in rec if x['arc_symbolsrt']])

        if count > 0:
            logger.debug(f'{count=}')
            return None

        if symbol_count == 0:
            logger.debug(f'{symbol_count=}')
            return None

    # Download xml file
    with tempfile.TemporaryDirectory() as tmpdir:
        xml_file = f'{tmpdir}/{record_id}-{event_id}-{raw_field}.xml'

        logger.debug(f'downloading file:{xml_file}')

        try:
            download_file(
                project,
                record_id,
                raw_field,
                xml_file,
                event_id=event_id)
        except Exception as err:
            logging.error(f'download failed:{record_id}:{event_id}:{err}')
            return None

        # Transform data
        try:
            data = _transform(xml_file)
            for d in data:
                d[project.def_field] = record_id
                d['redcap_event_name'] = event_id
        except Exception as err:
            logging.error(f'transform failed:{record_id}:{event_id}:{err}')
            import traceback
            traceback.print_exc()
            return None

    if not data:
        return None

    # Finally load back to redcap
    result = _load(project, data)

    return result


def _transform(filename):
    data = []

    # Load the data
    logging.info(f'loading:{filename}')
    file_data = extract_data(filename)

    file_data = [x for x in file_data if x['id'] > 0]

    for d in file_data:
        context1 = d.get('context1', '')
        context2 = d.get('context2', '')
        context3 = d.get('context3', '')
        context4 = d.get('context4', '')
        context5 = d.get('context5', '')

        if context1 == context2 == context3 == context4 == context4 == context5 == '':
            continue

        data.append({
            'redcap_repeat_instance': str(d['id']),
            'redcap_repeat_instrument': 'arc_data',
            'arc_context1': context1,
            'arc_context2': context2,
            'arc_context3': context3,
            'arc_context4': context4,
            'arc_context5': context5,
        })

    return data


def process_context_survey(data):
    result = {}

    try:
        result['context1'] = data['questions'][0]['text_value']
    except (KeyError, IndexError):
        result['context1'] = ''

    try:
        result['context2'] = data['questions'][1]['text_value']
    except (KeyError, IndexError):
        result['context2'] = ''

    try:
        result['context3'] = str(data['questions'][2]['value'])
    except (KeyError, IndexError):
        result['context3'] = ''

    try:
        result['context4'] = str(data['questions'][3]['value'])
    except (KeyError, IndexError):
        result['context4'] = ''

    try:
        result['context5'] = data['questions'][4]['text_value']
    except (KeyError, IndexError):
        result['context5'] = ''

    return result


def process_session(data):
    result = {}

    # Overview
    result['id'] = data['id']

    # Check if session has been completed
    if not data['finishedSession']:
        return result

    if not data['testData']:
        return {}

    # Context Survey
    test_data = data['testData'][0]
    context_data = process_context_survey(test_data['context_survey'])
    result.update(context_data)

    # Return the complete dictionary
    return result


def extract_data(filename):
    data = []
    jsondata = None

    root = etree.parse(filename).getroot()
    for node in root:
        if node.attrib['name'] == 'ParticipantState':
            _json = json.loads(node.text)
            jsondata = _json['testCycles'][0]['days']
            break

    # Process each day/session
    for day in jsondata:
        for sess in day['sessions']:
            sess_data = process_session(sess)
            if not sess_data:
                continue

            # Keep it
            data.append(sess_data)

    # Return the complete data
    return data



import logging
import os

import pandas as pd

from ....garjus import Garjus
from ..utils import file_age


logger = logging.getLogger('dashboard.issues.data')


# This is where we save our cache of the data
def get_filename():
    datadir = f'{Garjus.userdir()}/DATA'
    filename = f'{datadir}/issuesdata.pkl'

    try:
        os.makedirs(datadir)
    except FileExistsError:
        pass

    return filename


def get_data():
    g = Garjus()

    if not g.redcap_enabled():
        logger.debug('redcap not enabled, no issues data')
        return pd.DataFrame(columns=g.column_names('issues'))

    pid = g.redcap_pid()

    logger.info('loading issues')
    df = g.issues()

    # Sort by date and reset index
    df.sort_values(by=['DATETIME'], inplace=True, ascending=False)
    df.reset_index(inplace=True)

    df['STATUS'] = 'FAIL'
    df['LABEL'] = df['ID']

    df['SESSIONLINK'] = g.xnat_host()
    # + \
    #    '/data/projects/' + df['PROJECT'] + \
    #    '/subjects/' + df['SUBJECT'] + \
    #    '/experiments/' + df['SESSION']

    project2id = g.projects_setting(list(df.PROJECT.unique()), 'primary')

    df['PROJECTPID'] = df['PROJECT'].map(project2id)

    df['SUBJECTID'] = df['SUBJECT']

    # Load record IDs so we can link to the subject
    for p in df.PROJECT.unique():
        primary = g.primary(p)

        if not primary:
            logger.debug(f'no primary found:{p}')
            continue

        def_field = primary.def_field
        sec_field = primary.export_project_info()['secondary_unique_field']
        if sec_field:
            # Handle secondary ID
            rec = primary.export_records(fields=[def_field, sec_field])
            subj2id = {x[sec_field]: x[def_field] for x in rec if x[sec_field]}
            df.loc[df['PROJECT'] == p, 'SUBJECTID'] = df['SUBJECT'].map(
                subj2id)
        else:
            # ID is same as subject number for this project
            pass

    # Make project link
    df['PROJECTLINK'] = 'https://redcap.vanderbilt.edu/redcap_v14.2.2/' + \
        '/index.php?pid=' + df['PROJECTPID']

    # Make record link
    df['IDLINK'] = 'https://redcap.vanderbilt.edu/redcap_v14.2.2/' + \
        'DataEntry/index.php?' + \
        'pid=' + str(pid) + \
        '&page=issues&id=' + \
        df['PROJECT'] + \
        '&instance=' + \
        df['ID'].astype(str)

    # TODO: not able to work yet, need more redcap ids to get to page
    #  Make field link
    df['FIELDLINK'] = 'https://redcap.vanderbilt.edu/redcap_v14.2.2/' + \
        '/index.php?pid=' + df['PROJECTPID']

    return df


def run_refresh():
    filename = get_filename()

    df = get_data()

    if not df.empty:
        save_data(df, filename)

    return df


def load_data(refresh=False, maxmins=5):
    filename = get_filename()

    if not os.path.exists(filename):
        refresh = True
    elif file_age(filename) > maxmins:
        logger.info(f'refreshing, file age limit reached:{maxmins} minutes')
        refresh = True

    if refresh:
        df = run_refresh()
    else:
        df = read_data(filename)

    return df


def read_data(filename):

    if os.path.exists(filename):
        df = pd.read_pickle(filename)
    else:
        df = pd.DataFrame(columns=[
            'ID', 'LABEL', 'PROJECT', 'SUBJECT', 'SESSION',
            'EVENT', 'FIELD', 'CATEGORY', 'STATUS',
            'DESCRIPTION', 'DATETIME'
        ])

    return df


def save_data(df, filename):
    # save to cache
    df.to_pickle(filename)


def filter_data(df, projects, categories):
    # Filter by project
    if projects:
        logger.debug('filtering by project:')
        logger.debug(projects)
        df = df[df['PROJECT'].isin(projects)]

    # Filter by category
    if categories:
        logger.debug('filtering by category:')
        logger.debug(categories)
        df = df[(df['CATEGORY'].isin(categories))]

    return df



"""Life Data."""
import logging
import tempfile

import pandas as pd

from ...utils_redcap import field2events, download_file


logger = logging.getLogger('garjus.automations.etl_lifedata')


VARMAP = {
    'lifedata_others': 'Basic Info (2) (2)',
    'lifedata_worthless': 'Dep 1 (2) (2)',
    'lifedata_helpless': 'Dep 2 (2) (2)',
    'lifedata_depressed': 'Dep 3 (2) (2)',
    'lifedata_hopeless': 'Dep 4 (2) (2)',
    'lifedata_fatigued': 'Fatigue 1 (2) (2)',
    'lifedata_tired': 'Fatigue 2 (2) (2)',
    'lifedata_stress': 'Stress 1 (2) (2)',
    'lifedata_work': 'Stress 2 (2) (2)_Work',
    'lifedata_family': 'Stress 2 (2) (2)_Family',
    'lifedata_financial': 'Stress 2 (2) (2)_Financial',
    'lifedata_health': 'Stress 2 (2) (2)_Health',
    'lifedata_social': 'Stress 2 (2) (2)_Social (Non-family)',
    'lifedata_other': 'Stress 2 (2) (2)_Other',
    'lifedata_nostress': 'Stress 2 (2) (2)_No Stress',
    'lifedata_deserve': 'Rum 1 (2) (2)',
    'lifedata_react': 'Rum 2 (2) (2)',
    'lifedata_situation': 'Rum 3 (2) (2)',
    'lifedata_problems': 'Rum 4 (2) (2)',
    'lifedata_handle': 'Rum 5 (2) (2)',
    'lifedata_down': 'Pos/Neg 1 (2) (2)',
    'lifedata_happy': 'Pos/Neg 5 (2) (2)',
    'lifedata_guilty': 'Pos/Neg 2 (2) (2)',
    'lifedata_cheerful': 'Pos/Neg 6 (2) (2)',
    'lifedata_lonely': 'Pos/Neg 3 (2) (2)',
    'lifedata_satisfied': 'Pos/Neg 7 (2) (2)',
    'lifedata_anxious': 'Pos/Neg 4 (2) (2)',
}


SUMS = {
    'lifedata_deptot' : ['lifedata_worthless', 'lifedata_helpless', 'lifedata_depressed', 'lifedata_hopeless'],
    'lifedata_fatiguetot': ['lifedata_tired', 'lifedata_stress'],
    'lifedata_rumtot': ['lifedata_deserve', 'lifedata_react', 'lifedata_situation', 'lifedata_problems', 'lifedata_handle'],
    'lifedata_postot': ['lifedata_happy', 'lifedata_cheerful', 'lifedata_satisfied'],
    'lifedata_negtot': ['lifedata_down', 'lifedata_guilty', 'lifedata_lonely', 'lifedata_anxious']
}


logger = logging.getLogger('garjus.automations.etl_lifedata')


def sum_responses(data, labels):
    response_sum = 0
    for r in labels:
        if r not in data:
            return ''

        response_sum += int(data[r])

    return response_sum


def process_project(project):
    '''project is a pycap project for project that contains life data.'''
    file_field = 'life_file'
    results = []
    def_field = project.def_field
    fields = [def_field, file_field]
    id2subj = {}
    events = field2events(project, file_field)
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get records
    rec = project.export_records(fields=fields, events=events)

    # Process each record
    for r in rec:
        record_id = r[def_field]
        event_id = r['redcap_event_name']
        subj = id2subj.get(record_id)

        #if record_id != '3':
        #    continue

        # Check for converted file
        if not r[file_field]:
            logger.debug(f'no life file:{record_id}:{subj}:{event_id}')
            continue

        if r[file_field] == 'CONVERT_FAILED.txt':
            logger.debug(f'found CONVERT_FAILED')
            continue

        if r[file_field] == 'MISSING_DATA.txt':
            logger.debug(f'found MISSING_DATA')
            continue

        # Do the ETL
        result = _process(project, record_id, event_id)
        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_lifedata',
                'description': 'etl_lifedata',
                'subject': subj,
                'event': event_id,
                'field': file_field})

    return results


def _process(project, record_id, event_id):
    file_field = 'life_file'

    logger.debug(f'etl_lifedata:{record_id}:{event_id}')

    # check for existing records in the repeating instruments for this event
    rec = project.export_records(
        records=[record_id],
        events=[event_id],
        forms=['ema_lifedata_survey'],
        fields=[project.def_field]
    )

    rec = [x for x in rec if x['redcap_repeat_instrument'] == 'ema_lifedata_survey']

    if len(rec) > 0:
        logger.debug(f'found existing records:{record_id}:{event_id}')
        return

    with tempfile.TemporaryDirectory() as tmpdir:
        csv_file = f'{tmpdir}/{record_id}-{event_id}-{file_field}.csv'

        logger.debug(f'downloading file:{record_id}:{event_id}')

        try:
            download_file(
                project,
                record_id,
                file_field,
                csv_file,
                event_id=event_id)
        except Exception as err:
            logger.error(f'download failed:{record_id}:{event_id}:{err}')
            return False

        # Transform data
        try:
            data = _transform(csv_file)
            for d in data:
                d[project.def_field] = record_id
                d['redcap_event_name'] = event_id
        except Exception as err:
            logger.error(f'transform failed:{record_id}:{event_id}:{err}')
            import traceback
            traceback.print_exc()
            return False

    if not data:
        return False

    # Also complete the lifedata file form
    data.append({
        project.def_field: record_id,
        'redcap_event_name': event_id,
        'life_data_complete': '2'
    })

    # Finally load back to redcap
    result = _load(project, data)

    return result


def _load(project, data):
    # Load the data back to redcap, data list is dictionaries
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        return False

def _read(filename):
    df = pd.DataFrame()

    try:
        # Load Data
        df = pd.read_csv(filename, dtype=str)
    except ValueError as err:
        logger.error(f'failed to read excel:{filename}:{err}')

    return df


def get_response(df, label):
    if len(df[df['Prompt Label'] == label]) > 1:
        logger.debug('duplicates!')
    elif len(df[df['Prompt Label'] == label]) == 0:
        logger.debug(f'missing:{label}')
        return ''

    return df[df['Prompt Label'] == label].iloc[0].Response


def _transform(filename):
    data = []

    # Load the data
    logger.info(f'loading:{filename}')
    df = _read(filename)

    if df.empty:
        logger.debug(f'empty file')
        return []

    df = df.fillna('')

    if df is None:
        logger.error('extract failed')
        return

    # make a record per notification
    for i in df['Notification No'].unique():
        d = {
            'redcap_repeat_instance': str(i),
            'redcap_repeat_instrument': 'ema_lifedata_survey',
            'ema_lifedata_survey_complete': '2'}

        dfs = df[df['Notification No'] == i]

        if dfs.empty:
            logger.debug(f'no rows for Notification No:{i}')
            continue

        d['lifedata_notification_time'] = dfs.iloc[0]['Notification Time']
        d['lifedata_notification_no'] = dfs.iloc[0]['Notification No']

        if dfs.iloc[0]['Session Instance No']:
            d['lifedata_session_no'] = str(int(float(
                dfs.iloc[0]['Session Instance No'])))

        if dfs.iloc[0]['Responded'] != '1':
            d['lifedata_responded'] = '0'
        else:
            d['lifedata_responded'] = '1'

            if dfs.iloc[0]['Prompt Response Time']:
                # Get the response date and time separately
                d['lifedata_response_date'] = dfs.iloc[0]['Prompt Response Time'].split(' ')[0]
                d['lifedata_response_time'] = dfs.iloc[0]['Prompt Response Time'].split(' ')[1].rsplit(':', 1)[0]

                # Get the prompt values
                for k, v in VARMAP.items():
                    response = get_response(dfs, v)
                    if response:
                        d[k] = str(int(float(response)))

            if dfs.iloc[0]['Session Length']:
                d['lifedata_session_length'] = dfs.iloc[0]['Session Length'].split(':', 1)[1]

            # Calculate sums for this session
            for k, v in SUMS.items():
                # Creates a new value for key k in record d which sums
                # the values from the list of keys k
                d[k] = sum_responses(d, v)

        # Append to our list of records
        data.append(d)

    return data



'''Subjects from REDCap.'''
import logging

import pandas as pd
import numpy as np

from ..utils_redcap import secondary, secondary_map, field2events


# Here we provide an export of a limited set of fields as determined here, later we'll
# make those fields dynamic with entries in REDcap. The list of proctypes is 
# determined by projects included. Each proctype is merged across projects.
# each project is a separate database so we load each project separately


logger = logging.getLogger('garjus.subjects.covariates')


# TODO: move this data to redcap entries
TYPE2PROJECTS = {
    'examiner': ['D3', 'COGD'],
    'plasma': ['D3'],
    'gaitrite': ['D3'],
    'toolbox': ['D3'],
    'fallypride': ['D3'],
    'msit_vitals': ['REMBRANDT'],
}


def _load_toolbox(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [
        def_field,
        'toolbox_picseqtest_uncstd',
        'toolbox_listsorttest_uncstd',
        'toolbox_patterntest_uncstd',
        'toolbox_picvocabtest_uncstd',
        'toolbox_oralrecogtest_uncstd',
        'toolbox_cogcrystalcomp_uncstd'
    ]

    # TODO: get this dynamically, these are D3 events
    events = [
        'baseline_arm_1',
        'baseline_arm_2',
        'baseline_arm_3',
        'baseline_arm_4'
    ]

    rec = rc.export_records(fields=fields)

    # Filter by selected events
    rec = [x for x in rec if x['redcap_event_name'] in events]

    rec = [x for x in rec if x['toolbox_picseqtest_uncstd']]

    # Format as dataframe of strings dropping any dupes
    df = pd.DataFrame(rec, columns=fields)
    df = df.astype(str)
    df = df.drop_duplicates(subset=[def_field], keep='first')

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    df = df.rename(columns={
        'toolbox_picseqtest_uncstd': 'picseqtest_uncstd',
        'toolbox_listsorttest_uncstd': 'listsorttest_uncstd',
        'toolbox_patterntest_uncstd': 'patterntest_uncstd',
        'toolbox_picvocabtest_uncstd': 'picvocabtest_uncstd',
        'toolbox_oralrecogtest_uncstd': 'oralrecogtest_uncstd',
        'toolbox_cogcrystalcomp_uncstd': 'cogcrystalcomp_uncstd',
    })

    return df[[
        'ID',
        'picseqtest_uncstd',
        'listsorttest_uncstd',
        'patterntest_uncstd',
        'picvocabtest_uncstd',
        'oralrecogtest_uncstd',
        'cogcrystalcomp_uncstd'
        ]]


def _load_examiner(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [
        def_field,
        'executive_composite',
        'fluency_factor',
        'cog_control_factor',
        'working_memory_factor',
        'executive_se',
        'fluency_se',
        'cog_control_se',
        'working_memory_se',
    ]

    # TODO: get this dynamically, these are D3 events
    events = [
        'baseline_arm_1',
        'baseline_arm_2',
        'baseline_arm_3',
        'baseline_arm_4'
    ]

    rec = rc.export_records(fields=fields)

    # Filter by selected events
    rec = [x for x in rec if x['redcap_event_name'] in events]

    rec = [x for x in rec if x['executive_composite']]

    # Format as dataframe of strings dropping any dupes
    df = pd.DataFrame(rec, columns=fields)
    df = df.astype(str)
    df = df.drop_duplicates(subset=[def_field], keep='first')

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    return df[[
        'ID',
        'executive_composite',
        'fluency_factor',
        'cog_control_factor',
        'working_memory_factor',
        'executive_se',
        'fluency_se',
        'cog_control_se',
        'working_memory_se',
        ]]


def _load_gaitrite(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [
        def_field,
        'gaitrite_comments',
        'gaitrite_velocity',
    ]

    # TODO: get this dynamically, these are D3 events
    events = [
        'baseline_arm_1',
        'baseline_arm_2',
        'baseline_arm_3',
        'baseline_arm_4'
    ]

    rec = rc.export_records(fields=fields)

    # Only gaitrite records
    rec = [x for x in rec if x['redcap_repeat_instrument'] == 'gaitrite']

    # Filter by selected events
    rec = [x for x in rec if x['redcap_event_name'] in events]

    # Only specific walks
    rec = [x for x in rec if \
        x['gaitrite_comments'].lower().startswith('standard') or \
        x['gaitrite_comments'].lower().startswith('first')
    ]

    # Format as dataframe of strings dropping any dupes
    df = pd.DataFrame(rec, columns=fields)
    df = df.astype(str)
    df = df.drop_duplicates(subset=[def_field], keep='first')

    # Rename
    df['SPEED'] = df['gaitrite_velocity']

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    return df[['ID', 'SPEED']]


def _load_plasma(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [def_field, 'plasma_hscrp']
    events = field2events(rc, 'plasma_hscrp')

    rec = rc.export_records(fields=fields, events=events)

    # Filter by selected events
    # TODO: is this necessary?
    rec = [x for x in rec if x['redcap_event_name'] in events]

    rec = [x for x in rec if x['plasma_hscrp']]

    # Format as dataframe of strings dropping any dupes
    df = pd.DataFrame(rec, columns=fields)
    df = df.astype(str)
    df = df.drop_duplicates(subset=[def_field], keep='first')

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    return df[['ID','plasma_hscrp']]


def _load_fallypride(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [def_field, 'fallypride_accumbens', 'fallypride_amygdala', 'fallypride_caudate', 'fallypride_pallidum', 'fallypride_putamen', 'fallypride_thalamus']
    events = field2events(rc, 'fallypride_accumbens')

    rec = rc.export_records(fields=fields, events=events)
    rec = [x for x in rec if x['redcap_event_name'] in events]
    rec = [x for x in rec if x['fallypride_accumbens']]

    # Format as dataframe of strings dropping any dupes
    df = pd.DataFrame(rec, columns=fields)
    df = df.astype(str)
    df = df.drop_duplicates(subset=[def_field], keep='first')

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    return df[['ID', 'fallypride_accumbens', 'fallypride_amygdala', 'fallypride_caudate', 'fallypride_pallidum', 'fallypride_putamen', 'fallypride_thalamus']]


def _load_msit_vitals(rc):
    df = pd.DataFrame()
    def_field = rc.def_field
    fields = [
        def_field,
        't1_pulse2', 't1_pulse3',
        't1_sys2', 't1_sys3',
        't1_dia2', 't1_dia3',
        'msit_pulse1',  'msit_pulse2',  'msit_pulse3',  'msit_pulse4',
        'msit_dia1', 'msit_dia2', 'msit_dia3', 'msit_dia4',
        'msit_sys1', 'msit_sys2', 'msit_sys3', 'msit_sys4',
    ]
    events = field2events(rc, 'msit_pulse1')

    # TODO: get this dynamically, these are REMBRANDT events
    events = [
        'baselinemonth_0_arm_2',
        'baselinemonth_0_arm_3',
    ]

    rec = rc.export_records(fields=fields, events=events)
    rec = [x for x in rec if x['redcap_event_name'] in events]
    rec = [x for x in rec if x['msit_pulse1']]

    df = pd.DataFrame(rec, columns=fields + ['redcap_event_name'])

    # Set subject
    sec_field = secondary(rc)
    if sec_field:
        # Get the secondary field values to set the subject ID
        sec_map = secondary_map(rc)
        df['ID'] = df[def_field].map(sec_map)
    else:
        df['ID'] = df[def_field]

    df = df.replace('', np.nan).dropna(how='any', axis=0)

    df['rest_dbp'] = (df['t1_dia2'].astype('float') + df['t1_dia3'].astype('float')) / 2
    df['rest_pulse'] = (df['t1_pulse2'].astype('float') + df['t1_pulse3'].astype('float')) / 2
    df['rest_sbp'] = (df['t1_sys2'].astype('float') + df['t1_sys3'].astype('float')) / 2

    df['msit_dbp'] = (
        df['msit_dia1'].astype('float') + \
        df['msit_dia2'].astype('float') + \
        df['msit_dia3'].astype('float') + \
        df['msit_dia4'].astype('float')) / 4
    
    df['msit_pulse'] = (
        df['msit_pulse1'].astype('float') + \
        df['msit_pulse2'].astype('float') + \
        df['msit_pulse3'].astype('float') + \
        df['msit_pulse4'].astype('float')) / 4

    df['msit_sbp'] = (
        df['msit_sys1'].astype('float') + \
        df['msit_sys2'].astype('float') + \
        df['msit_sys3'].astype('float') + \
        df['msit_sys4'].astype('float')) / 4

    df['chg_pulse'] = df['msit_pulse'] - df['rest_pulse']
    df['chg_sbp'] = df['msit_sbp'] - df['rest_sbp']

    df = df.sort_values(['ID', 'redcap_event_name'])

    df.drop('record_id', axis=1)

    return df


def _proctype_projects(proctype, projects):
    proctype_projects = TYPE2PROJECTS.get(proctype, [])
    return list(set(projects) & set(proctype_projects))


def _export_toolbox(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()

    projects = _proctype_projects('toolbox', subjects_df.PROJECT.unique())

    logger.info(f'loading toolbox for projects:{projects}')

    for p in projects:
        # Load data for project
        logger.info(f'loading toolbox:{p}')
        _df = _load_toolbox(garjus.primary(p))

        # Filter by subjects
        logger.info(f'filtering by subject')
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/toolbox.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df= df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info('no toolbox data to save')

    return df


def _export_examiner(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()

    projects = _proctype_projects('examiner', subjects_df.PROJECT.unique())

    logger.info(f'loading examiner for projects:{projects}')

    for p in projects:
        # Load examiner data for project
        logger.info(f'loading examiner:{p}')
        _df = _load_examiner(garjus.primary(p))

        # Filter by subjects
        logger.info(f'filtering by subject')
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/examiner.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df= df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info('no examiner data to save')

    return df


def _export_gaitrite(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()

    projects = _proctype_projects('gaitrite', subjects_df.PROJECT.unique())

    logger.info(f'loading gaitrite for projects:{projects}')

    for p in projects:
        # Load gaitrite data for project
        logger.info(f'loading gaitrite:{p}')
        _df = _load_gaitrite(garjus.primary(p))

        # Filter by subjects
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/gaitrite.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df = df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info('no gaitrite data to save')

    return df


def _export_plasma(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()

    projects = _proctype_projects('plasma', subjects_df.PROJECT.unique())

    logger.info(f'loading plasma for projects:{projects}')

    for p in projects:
        # Load plasma data for project
        logger.info(f'loading plasma:{p}')
        _df = _load_plasma(garjus.primary(p))

        # Filter by subjects
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/plasma.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df = df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info('no plasma data to save')

    return df


def _export_fallypride(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()
    proctype = 'fallypride'
    projects = _proctype_projects(proctype, subjects_df.PROJECT.unique())

    logger.info(f'loading {proctype} for projects:{projects}')

    for p in projects:
        # Load data for project
        logger.info(f'loading {proctype}:{p}')
        _df = _load_fallypride(garjus.primary(p))

        # Filter by subjects
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/{proctype}.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df = df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info(f'no {proctype} data to save')

    return df


def _export_msit_vitals(garjus, tmpdir, subjects_df):
    df = pd.DataFrame()
    proctype = 'msit_vitals'
    projects = _proctype_projects(proctype, subjects_df.PROJECT.unique())

    logger.info(f'loading {proctype} for projects:{projects}')

    for p in projects:
        # Load data for project
        logger.info(f'loading {proctype}:{p}')
        _df = _load_msit_vitals(garjus.primary(p))

        # Filter by subjects
        subjects = subjects_df[subjects_df['PROJECT'] == p].ID.unique()
        _df = _df[_df.ID.isin(subjects)]

        # Append to whole
        _df['PROJECT'] = p
        df = pd.concat([df, _df], ignore_index=True)

    if len(df) > 0:
        # Save to csv
        _file = f'{tmpdir}/{proctype}.csv'
        logger.info(f'saving to csv:{_file}')
        _cols = ['ID', 'PROJECT'] + [x for x in df.columns if x not in ['ID', 'PROJECT']]
        df = df.sort_values(['PROJECT', 'ID'])
        df.to_csv(_file, index=False, columns=_cols)
    else:
        logger.info(f'no {proctype} data to save')

    return df



def export_clinical():
    df = pd.DataFrame()

    # Load D3 clinical data

    # MADRS baseline
    # ma_tot
    # madrs_date

    # ACS baseline
    # acs_focus
    # acs_shift
    # acs_sum
    # acs_date

    # QIDS
    # qidsc_total
    # qids_date

    # AES
    # aes_total
    # aes_date
    # aes_totalc
    # aes_totalb
    # aes_totale
    # aes_totalo

    return df


def export(garjus, tmpdir, subjects_df):
    data = {}

    # Save csv for each type filtered by subject list
    _data = _export_examiner(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        _data.loc[_data.ID.str.startswith('P'), 'SITE'] = 'UPMC'
        data['examiner'] = _data

    _data = _export_gaitrite(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        _data.loc[_data.ID.str.startswith('P'), 'SITE'] = 'UPMC'
        data['gaitrite'] = _data 

    _data = _export_plasma(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        _data.loc[_data.ID.str.startswith('P'), 'SITE'] = 'UPMC'
        data['plasma'] = _data

    _data = _export_toolbox(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        _data.loc[_data.ID.str.startswith('P'), 'SITE'] = 'UPMC'
        data['toolbox'] = _data

    _data = _export_fallypride(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        data['fallypride'] = _data

    _data = _export_msit_vitals(garjus, tmpdir, subjects_df)
    if not _data.empty:
        _data['SITE'] = 'VUMC'
        _data.loc[_data.ID.str.startswith('2'), 'SITE'] = 'UPMC'
        _data.loc[_data.ID.str.startswith('3'), 'SITE'] = 'UIC'
        data['msit_vitals'] = _data

    return data



"""ARC app."""
import logging
import tempfile
import json
import math
from datetime import datetime

import pandas as pd

from ...utils_redcap import download_named_file
from . import file2redcap
from . import summary


logger = logging.getLogger('garjus.automations.etl_arcdata')


file_field = 'arc_testfile'


def process(project, datadir):
    results = []
    def_field = project.def_field
    fields = [def_field, file_field, 'arc_missedsession']

    logger.debug('file2redcap')
    results = file2redcap.process(project, datadir)

    logger.debug('process')

    # Handle secondary ID
    sec_field = project.export_project_info()['secondary_unique_field']
    if sec_field:
        rec = project.export_records(fields=[def_field, sec_field])
        id2subj = {x[def_field]: x[sec_field] for x in rec if x[sec_field]}
    else:
        rec = project.export_records(fields=[def_field])
        id2subj = {x[def_field]: x[def_field] for x in rec if x[def_field]}

    # Get file records
    logger.debug('getting arc testfile records')
    records = project.export_records(fields=fields)
    records = [x for x in records if x['redcap_repeat_instance']]
    records = [x for x in records if x[file_field]]

    logger.debug('processing arc testfile records')

    for r in records:

        record_id = r[def_field]
        event_id = r['redcap_event_name']
        repeat_id = r['redcap_repeat_instance']

        if event_id.startswith('unscheduledad_hoc_arm_3') or event_id.startswith('screening'):
            logger.debug(f'skipping:{record_id}:{event_id}:{repeat_id}')
            continue

        if r['arc_missedsession']:
            logger.debug(f'already extracted:{record_id}:{event_id}:{repeat_id}')
            continue

        # Process record
        logger.info(f'process:{r[file_field]}:{record_id}:{event_id}:{repeat_id}')
        result = _process(project, record_id, event_id, repeat_id)

        if result:
            results.append({
                'result': 'COMPLETE',
                'category': 'etl_arcdata',
                'description': 'etl_arcdata',
                'subject': id2subj[record_id],
                'event': event_id,
                'repeat': repeat_id,
                'field': file_field})

    logger.debug('summary')
    results = results + summary.process(project)

    return results


def _process(project, record_id, event_id, repeat_id):
    test_file = None
    result = []
    data = {}

    with tempfile.TemporaryDirectory() as tmpdir:
        logger.debug(f'downloading file:{record_id}:{event_id}:{repeat_id}') 

        try:
            test_file = download_named_file(
                project,
                record_id,
                file_field,
                tmpdir,
                event_id=event_id,
                repeat_id=repeat_id
            )
        except Exception as err:
            logging.error(f'download failed:{record_id}:{event_id}:{repeat_id}:{err}')
            return False

        logger.debug(f'transforming:{record_id}:{event_id}:{repeat_id}') 
        try:
            data = _transform(test_file)
        except Exception as err:
            logging.error(f'transform failed:{record_id}:{event_id}:{repeat_id}:{err}')
            import traceback
            traceback.print_exc()
            return False

    #if 'arc_context1' not in data:
    #    logging.error(f'nothing found:{record_id}:{event_id}:{repeat_id}')
    #    return False

    data[project.def_field] = record_id
    data['redcap_event_name'] = event_id
    data['redcap_repeat_instrument'] = 'arc_data'
    data['redcap_repeat_instance'] = repeat_id

    # Finally load back to redcap
    result = _load(project, [data])

    return result


def _process_context_survey(data):

    try:
        return {
            'arc_context1': data['questions'][0]['text_value'],
            'arc_context2': data['questions'][1]['text_value'],
            'arc_context3': str(data['questions'][2]['value']),
            'arc_context4': str(data['questions'][3]['value']),
            'arc_context5': data['questions'][4]['text_value'],
        }
    except (KeyError, IndexError):
        return {}


def _process_price_test(data):
    df = pd.DataFrame(data['sections'])

    if 'selection_time' not in df.columns:
        return {}

    # Calculate accuracy
    df['acc'] = (df['selected_index'] == df['correct_index'])
    price_acc = (df['acc'].mean() * 100).round().astype(int)

    # Calculate RT with correct trials
    df['rt'] = df['selection_time'] - df['stimulus_display_time']
    price_rt = df[df['acc'] == 1]['rt'].median()

    return {
        'arc_prices_accuracy': str(price_acc),
        'arc_pricesrt': str(round(price_rt, 3))
    }


def _process_symbol_test(data):
    df = pd.DataFrame(data['sections'])

    if 'selection_time' not in df.columns:
        return {}

    # Calculate accuracy
    df['acc'] = (df['selected'] == df['correct'])
    symbol_acc = (df['acc'].mean() * 100).round().astype(int)

    # Calculate RT with correct trials
    df['rt'] = df['selection_time'] - df['appearance_time']
    symbol_rt = df[df['acc'] == 1]['rt'].median()
    return {
        'arc_symbols_accuracy': str(symbol_acc),
        'arc_symbolsrt': str(round(symbol_rt, 3)),
    }


def _process_grid_test(data):
    ed_sec = []
    for s, cur_section in enumerate(data['sections']):
        if len(cur_section['choices']) > 0:
            for c, cur_choice in enumerate(cur_section['choices']):
                cur_ed = [None] * len(cur_section['images'])
                for i, cur_image in enumerate(cur_section['images']):
                    ix = cur_image['x']
                    iy = cur_image['y']
                    cx = cur_choice['x']
                    cy = cur_choice['y']
                    cur_ed[i] = math.sqrt(
                        ((cx - ix) * (cx - ix)) + ((cy - iy) * (cy - iy))
                    )

            # Keep running sum of Mininum ED for each choice
            ed_sec.append(min(cur_ed))
        else:
            logger.debug('empty choices')

    if not ed_sec:
        return {}
    else:
        # Average ED of sections
        grid_ed = sum(ed_sec) / float(len(ed_sec))
        return {
            'arc_grided': str(round(grid_ed, 3))
        }


def _load_testfile(testfile):
    data = []

    with open(testfile) as f:
        data = json.load(f)

    return data


def _load(project, data):
    # Load the data back to redcap
    try:
        _response = project.import_records(data)
        assert 'count' in _response
        return True
    except (AssertionError, Exception) as err:
        logger.error(err)
        return False


def _transform(filename):
    data = {}
    file_data = _load_testfile(filename)

    repeat_id = str(file_data['id'])
    data = {
        'redcap_repeat_instance': repeat_id,
        'redcap_repeat_instrument': 'arc_data',
        'arc_order_index': repeat_id,
        'arc_day_index': str(file_data['day']),
        'arc_data_complete': '2',
        'arc_interrupted': str(file_data['interrupted']),
        'arc_response_date': datetime.utcfromtimestamp(
            file_data['session_date']).strftime('%Y-%m-%d'),
        'arc_notification_time': datetime.fromtimestamp(
            file_data['session_date']).strftime('%H:%M'),
        'arc_finishedsession': str(file_data['finished_session']),
        'arc_missedsession': str(file_data['missed_session']),
        'arc_index_no': str(file_data['session']),
        'arc_session_no': str(file_data['session_id']),
    }

    if 'start_time' in file_data:
        data['arc_starttime'] = datetime.fromtimestamp(
            file_data['start_time']).strftime('%H:%M')

    if len(file_data.get('tests', 0)) > 0:
        test_data = file_data['tests'][0]

        if test_data.get('context_survey', False):
            data.update(_process_context_survey(test_data['context_survey']))

        if test_data.get('symbol_test', False):
            data.update(_process_symbol_test(test_data['symbol_test']))

        if test_data.get('price_test', False):
            data.update(_process_price_test(test_data['price_test']))

        if test_data.get('grid_test', False):
            data.update(_process_grid_test(test_data['grid_test']))

    return data
